/* Copyright 2018 Kevin Ryde

   This file is part of nautyextra.

   nautyextra is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3, or (at your option) any later
   version.

   nautyextra is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
   more details.

   You should have received a copy of the GNU General Public License along
   with nautyextra.  See the file COPYING.  If not, see
   <http://www.gnu.org/licenses/>.
*/

/* Search for asymmetric graphs which can be made symmetric by deleting or
   adding one edge.  Or rather can be by delete and cannot be by adding.

   cf A124059 num connected asymmetric


   --------
   can by delete, cannot by add

   >>graph6<<HCrfRjU
   >>graph6<<HCZJerr
   n=9 found 2

   n=10 found 206

   --------
   cannot by delete, can by add
   graph complements

   >>graph6<<H?qbtre
   >>graph6<<HCQfeql
   >Z 261080 graphs generated in 3.78 sec
   n=9 found 2

   n=9
   add 2--3
   .       0  1  2  3  4  5  6  7  8
   autom   0  1  2  3  4  5  6  7  8
   autom   4  3  8  5  7  1  2  0  6
   autom   7  5  6  1  0  3  8  4  2

   n=9
   add 1--2

   -------------
   n=9 automorphism group size 3
   >>graph6<<HCOedLj
   >>graph6<<HCpdmji
   >>graph6<<HCrVNZr
   >>graph6<<HCZJerb

   n=10 automorphism group size 3
*/

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <nauty/nauty.h>
#include <nauty/naugroup.h>
#include <nauty/nausparse.h>
#include <nauty/naututil.h>
#include <nauty/gtools.h>
#include <nauty/gutils.h>

#include "nautyextra.h"
#include "nautyextra-wip.h"

#define WANT_DEBUG  1  /* set to 1 for debug prints */
#if WANT_DEBUG
#define DEBUG(expr)  expr
#else
#define DEBUG(expr)
#endif


int option_min_vertices = 9;
int option_max_vertices = 9;
int option_verbose = 0;
int option_view = 0;
int option_count_found_limit = 30;
int option_view_found_limit = 5;

int my_geng_main(int argc, char**argv);

int count_total = 0;
int count_found = 0;

void
writeautom (int *p, int n)
{
  int i;
  printf("autom ");
  for (i = 0; i < n; i++)
    printf(" %2d",p[i]);
  printf("\n");
}
int countautom_total;
void
countautom (int *p, int n)
{
  countautom_total++;
}

/* g,m,n is an undirected graph.
   Return TRUE if it can be made symmetric by deleting one edge.
*/
int
is_symmetric_on_delete_edge1 (const graph *g, int m, int n)
{
  if (option_verbose)
    printf("is_symmetric_on_delete_edge1() n=%d\n", n);

  int from, to;
  int ret = FALSE;
  for (from = 0; from < n; from++) {
    set *from_row = GRAPHROW(g,from,m);
    for (to = from+1; to < n; to++) {
      if (ISELEMENT(from_row,to)) {
        graph h[MAXN*SETWORDSNEEDED(MAXN)];
        memcpy(h,g, n*sizeof(setword));
        DELONEEDGE(h, from,to, m);
        if (! nautyextra_is_asymmetric(h,m,n)) {
          if (option_verbose) {
            graph hc[MAXN*SETWORDSNEEDED(MAXN)];
            nautyextra_canonise(h,m,n, hc);
            printf(" symmetric on delete %d %d: to (canonical) %s%s",
                   from, to, GRAPH6_HEADER,ntog6(hc,m,n));
            ret = TRUE;
          } else {
            return TRUE;
          }
        }
      }
    }
  }
  return ret;
}

/* g,m,n is an undirected graph.
   Return TRUE if it can be made symmetric by adding one edge.
*/
int
is_symmetric_on_add_edge1 (const graph *g, int m, int n)
{
  if (option_verbose)
    printf("is_symmetric_on_add_edge1() n=%d\n", n);

  int ret = FALSE;
  int from, to;
  for (from = 0; from < n; from++) {
    set *from_row = GRAPHROW(g,from,m);
    for (to = from+1; to < n; to++) {
      if (! ISELEMENT(from_row,to)) {
        graph h[MAXN*SETWORDSNEEDED(MAXN)];
        memcpy(h,g, n*sizeof(setword));
        ADDONEEDGE(h, from,to, m);
        if (! nautyextra_is_asymmetric(h,m,n)) {
          if (option_verbose) {
            ret = TRUE;
            printf(" symmetric on add %d %d\n", from, to);
          } else {
            return TRUE;
          }
        }
      }
    }
  }
  return ret;
}
int
is_asymmetric_on_add_edge1 (const graph *g, int m, int n)
{
  if (option_verbose)
    printf("is_asymmetric_on_add_edge1() n=%d\n", n);

  int ret = FALSE;
  int from, to;
  for (from = 0; from < n; from++) {
    set *from_row = GRAPHROW(g,from,m);
    for (to = from+1; to < n; to++) {
      if (! ISELEMENT(from_row,to)) {
        graph h[MAXN*SETWORDSNEEDED(MAXN)];
        memcpy(h,g, n*sizeof(setword));
        ADDONEEDGE(h, from,to, m);
        if (nautyextra_is_asymmetric(h,m,n)) {
          if (option_verbose) {
            ret = TRUE;
            printf(" asymmetric on add %d %d\n", from, to);
            /* graph hc[MAXN*SETWORDSNEEDED(MAXN)]; */
            /* nautyextra_canonise(h,m,n, hc); */
            /* printf("%s%s", GRAPH6_HEADER,ntog6(hc,m,n)); */
          } else {
            return TRUE;
          }
        }
      }
    }
  }
  return ret;
}

void
show_with_add(const graph *g_orig, int m, int n, int u, int v)
{
  assert(m==1);
  graph g[MAXN];
  memcpy(g, g_orig, n*sizeof(setword));
  printf(" show add %d %d  %s%s", u,v, GRAPH6_HEADER,ntog6(g,m,n));
  ADDONEEDGE(g, u,v, m);

  printf(" show add %d %d  %s%s", u,v, GRAPH6_HEADER,ntog6(g,m,n));

  static DEFAULTOPTIONS_DENSEGRAPH(options);
  options.userautomproc = groupautomproc;
  options.userlevelproc = grouplevelproc;
  int lab[MAXN];
  int ptn[MAXN];
  int orbits[MAXN];
  statsblk stats;

  densenauty (g, lab, ptn, orbits, &options, &stats, m,n, NULL);
  printf("orig numorbits %d\n", stats.numorbits);

  printf ("automorphism group size = ");
  writegroupsize(stdout, stats.grpsize1, stats.grpsize2);
  printf ("\n");
  {
    grouprec *group = groupptr(FALSE);
    if (! group) {
      printf("no grouprec\n");
    } else {
      printf("grouprec %p\n", group);
      makecosetreps(group);  /* ready for allgroup() */
      allgroup(group,writeautom);
    }
  }
}

int
num_automorphisms(const graph *g, int m, int n)
{
  static DEFAULTOPTIONS_DENSEGRAPH(options);
  options.userautomproc = groupautomproc;
  options.userlevelproc = grouplevelproc;
  int lab[MAXN];
  int ptn[MAXN];
  int orbits[MAXN];
  statsblk stats;

  densenauty ((graph*)g, lab, ptn, orbits, &options, &stats, m,n, NULL);
  /* printf("numorbits %d\n", stats.numorbits); */

  if (stats.grpsize2 == 0) {
    return stats.grpsize1;
  }
  abort();

  /* writegroupsize(stdout, stats.grpsize1, stats.grpsize2); */
  /* grouprec *group = groupptr(FALSE); */
  /* if (! group) { */
  /*   printf("no grouprec\n"); */
  /*   abort(); */
  /* } */
  /* countautom_total = 0; */
  /* allgroup(group,countautom); */
  /* return countautom_total; */
}

void
my_geng_outproc (FILE *outfile, graph *g, int n)
{
  int m = SETWORDSNEEDED(n);
  assert (m == 1);

  if (0 && nautyextra_is_asymmetric(g,m,n)) {
    count_found++;
    printf("%s%s",GRAPH6_HEADER,ntog6(g,m,n));
  }
  if (0 && num_automorphisms(g,m,n) == 3) {
    count_found++;
    printf("n=%d e=%d\n", n, nautyextra_num_edges(g,m,n));
    printf("  %s%s",GRAPH6_HEADER,ntog6(g,m,n));
  }
  if (0 && nautyextra_regular_degree(g,m,n) == 4 && num_automorphisms(g,m,n)%3==0) {
    count_found++;
    printf("n=%d e=%d autos=%d\n",
           n, nautyextra_num_edges(g,m,n), num_automorphisms(g,m,n));
    printf("  %s%s",GRAPH6_HEADER,ntog6(g,m,n));
  }
  if (1
      && nautyextra_is_asymmetric(g,m,n)
      && is_symmetric_on_delete_edge1(g,m,n)
      && ! is_symmetric_on_add_edge1(g,m,n)) {

    count_found++;
    printf("found n=%d can del, cannot add\n", n);
    printf("%s%s",GRAPH6_HEADER,ntog6(g,m,n));
    /* nautyextra_print_matrix(g,m,n); */

    option_verbose = 1;
    is_symmetric_on_delete_edge1(g,m,n);
    is_symmetric_on_add_edge1(g,m,n);
    show_with_add(g,m,n, 2,3);
    option_verbose = 0;

    if (n < 7 || count_found < 6) {
      if (option_view) nautyextra_view(g,m,n);
    }
  }
}

void
try_Erdos_and_Renyi(void)
{
  /* Erdos and Renyi, "Asymmetric Graphs", Acta Mathematica Academiae
     Scientiarum Hungaricae, volume 14, 1963, pages 295-315
     https://users.renyi.hu/~p_erdos/1963-04.pdf

     .
     .  0
     .
     .     1     2
     .
     .  3     4     5    6
     .
     .     7     8
     .
     .  9

     whole
     >>graph6<<Ir[gGOzCO
     n=9 $=0 g3,5,7,8; 1:4,5,6,7,8; 2:4,6,8; 3:6,7,8; 4:5,7; 6:8; 7:8.

     delete 8,9
     >>graph6<<Ir[gGOzCO
     n=9 $=0 g3,5,7,8; 1:4,5,6,7,8; 2:4,6,8; 3:6,7; 4:5,7; 6:8; 7:8.

     Erdos and Renyi delete 8,9 makes 3-rotational symmetry.

     Also delete 4,5 makes an involution.  They note a delete to
     involution means can add to involution too, which is add 3,5.

     4,5 delete
     autom: 0 1 2  3 4  5 6  7 8 9
     autom: 0 1 2  4 3  5 6  9 8 7      swap 3,4 and 7,9

  */
  option_verbose = 1;
  static graph g[MAXN];
  static graph orig[MAXN];
  const int n = 10;
  int m = SETWORDSNEEDED(n);
  ADDONEEDGE(g, 0,1, m); ADDONEEDGE(g, 0,2, m);
  ADDONEEDGE(g, 1,3, m); ADDONEEDGE(g, 1,4, m);
  ADDONEEDGE(g, 2,3, m); ADDONEEDGE(g, 2,4, m); ADDONEEDGE(g, 2,5, m);
  ADDONEEDGE(g, 2,8, m);
  ADDONEEDGE(g, 3,4, m); ADDONEEDGE(g, 3,8, m); ADDONEEDGE(g, 3,9, m);
  ADDONEEDGE(g, 4,5, m); ADDONEEDGE(g, 4,7, m); ADDONEEDGE(g, 4,8, m);
  ADDONEEDGE(g, 5,6, m);
  ADDONEEDGE(g, 6,8, m);
  ADDONEEDGE(g, 7,8, m); ADDONEEDGE(g, 7,9, m);
  ADDONEEDGE(g, 8,9, m);
  memcpy(orig,g,sizeof(g));
  printf("whole graph:  %s%s",GRAPH6_HEADER,ntog6(g,m,n));
  printf("sizeof %d %d\n", sizeof(g), sizeof(orig));

  printf ("asymmetric  %d\n", nautyextra_is_asymmetric(g,m,n));
  printf ("symmetric on del  %d\n", is_symmetric_on_delete_edge1(g,m,n));
  printf ("symmetric on add  %d\n", is_symmetric_on_add_edge1(g,m,n));

  assert(nautyextra_is_asymmetric(g,m,n));
  assert(is_symmetric_on_delete_edge1(g,m,n));
  assert(is_symmetric_on_add_edge1(g,m,n));

  /* 8-9 delete as given by Erdos and Renyi */
  DELONEEDGE(g, 8,9, m);
  printf ("del asymmetric  %d\n", nautyextra_is_asymmetric(g,m,n));
  assert(nautyextra_is_asymmetric(g,m,n) == 0);  /* becomes symmetric */
  ADDONEEDGE(g, 8,9, m);
  assert(memcmp(orig,g,sizeof(g))==0);

  /* 4-5 other delete */
  DELONEEDGE(g, 4,5, m);
  printf ("other del asymmetric  %d\n", nautyextra_is_asymmetric(g,m,n));
  assert(nautyextra_is_asymmetric(g,m,n) == 0);  /* becomes symmetric */
  ADDONEEDGE(g, 4,5, m);
  assert(memcmp(orig,g,sizeof(g))==0);

  /* 5-3 other add */
  ADDONEEDGE(g, 5,3, m);
  printf ("other del asymmetric  %d\n", nautyextra_is_asymmetric(g,m,n));
  assert(nautyextra_is_asymmetric(g,m,n) == 0);  /* becomes symmetric */
  DELONEEDGE(g, 5,3, m);
  assert(memcmp(orig,g,sizeof(g))==0);

  static DEFAULTOPTIONS_DENSEGRAPH(options);
  options.userautomproc = groupautomproc;
  options.userlevelproc = grouplevelproc;
  int lab[MAXN];
  int ptn[MAXN];
  int orbits[MAXN];
  statsblk stats;

  densenauty (g, lab, ptn, orbits, &options, &stats, m,n, NULL);
  printf("orig numorbits %d\n", stats.numorbits);

  /* DELONEEDGE(g, 8,9, m); */
  ADDONEEDGE(g, 5,3, m);
  /* DELONEEDGE(g, 4,5, m); */

  printf("modified graph:  %s%s",GRAPH6_HEADER,ntog6(g,m,n));
  densenauty (g, lab, ptn, orbits, &options, &stats, m,n, NULL);
  printf("add numorbits %d\n", stats.numorbits);
  printf ("  orbits      ");
  int i;
  for (i = 0; i < n; i++) printf(" %d",orbits[i]);
  printf("\n");
  printf ("  v =         ");
  for (i = 0; i < n; i++) printf(" %d",i);
  printf("\n");

  int degrees[MAXN];
  nautyextra_degrees_array(g,m,n, degrees);
  printf ("deg =         ");
  for (i = 0; i < n; i++) printf(" %d",degrees[i]);
  printf("\n");

  printf ("automorphism group size = ");
  writegroupsize(stdout, stats.grpsize1, stats.grpsize2);
  printf ("\n");
  {
    grouprec *group = groupptr(FALSE);
    if (! group) {
      printf("no grouprec\n");
    } else {
      printf("grouprec %p\n", group);
      makecosetreps(group);  /* ready for allgroup() */
      allgroup(group,writeautom);
    }
  }

  /* exit (0); */
  option_verbose = 0;

  printf("-------------------\n");
}

int
main (void)
{
  setbuf(stdout,NULL);

  if (! getenv("DISPLAY"))
    option_view = 0;

  /* printf("NAUTYVERSIONID = %d\n", NAUTYVERSIONID); */

  if (0)
    try_Erdos_and_Renyi();

  if (0) {
    printf("\n");
    printf("n=9 after delete\n");
    const char str[] = "HCrbRjU";
    option_verbose = 1;
    static graph g[MAXN];
    const int n = 9;
    int m = SETWORDSNEEDED(n);
    stringtograph((char*) str,g,m);
    printf ("symmetric on add  %d\n", is_asymmetric_on_add_edge1(g,m,n));

    printf("\n");
    option_verbose = 0;
  }

  if (option_max_vertices < option_min_vertices) {
    printf ("max_vertices < min_vertices\n");
    abort();
  }
#if MAXN
  if (option_max_vertices > MAXN) {
    printf ("max_vertices %d bigger than MAXN %d\n", option_max_vertices, MAXN);
    abort();
  }
#endif

  int  num_vertices;
  for (num_vertices = option_min_vertices;
       num_vertices <= option_max_vertices;
       num_vertices++) {
    printf("\nn=%d\n", num_vertices);
    count_found = 0;

    {
      int n = num_vertices;
      int m = SETWORDSNEEDED(n);
      nauty_check(WORDSIZE,m,n,NAUTYVERSIONID);
    }

    int geng_argc = 0;
    char *geng_argv[10];
    geng_argv[geng_argc++] = "nauty-geng";
    if (0) geng_argv[geng_argc++] = "-q";     /* quiet */
    if (1) geng_argv[geng_argc++] = "-c";     /* connected graphs */

    char num_vertices_str[20];
    sprintf(num_vertices_str, "%d", num_vertices);
    geng_argv[geng_argc++] = num_vertices_str;

    /* 9*8/2 == 36 */
    /* int max_edges = ( num_vertices * (num_vertices-1) ) >> 1; */
    /* max_edges = (max_edges+1) >> 1; */

    int max_edges = 0;
    /* max_edges = 6; */
    char edges_range_str[20];
    if (max_edges) {
      sprintf(edges_range_str, "1:%d", max_edges);
      geng_argv[geng_argc++] = edges_range_str;
    }

    geng_argv[geng_argc] = NULL;

    printf("geng\n");
    my_geng_main(geng_argc,geng_argv);

    printf("n=%d found %d\n", num_vertices, count_found);
    if (count_found > 20) break;
  }
  printf("count total %d found %d\n", count_total, count_found);

  return 0;
}
