/* Copyright 2017, 2018, 2022 Kevin Ryde

   This file is part of nautyextra.

   nautyextra is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3, or (at your option) any later
   version.

   nautyextra is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
   more details.

   You should have received a copy of the GNU General Public License along
   with nautyextra.  See the file COPYING.  If not, see
   <http://www.gnu.org/licenses/>.
*/

/* Run some tests on nautyextra.c.
 */

#include <assert.h>
#include <math.h>
#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <nauty/nauty.h>
#include <nauty/nausparse.h>
#include <nauty/naututil.h>
#include <nauty/gtools.h>
#include <nauty/gutils.h>

#include "nautyextra.h"
#include "nautyextra-wip.h"

#define WANT_DEBUG  0  /* set to 1 for debug prints */
#if WANT_DEBUG
#define DEBUG(expr)  expr
#else
#define DEBUG(expr)
#endif

int option_verbose = 0;

const int tree_min_vertices = 2;
const int tree_max_vertices = 14;

const int graph_min_vertices = 1;
const int graph_max_vertices = 9;
const int ungraph_min_vertices = 1;
const int ungraph_max_vertices = 8;


int count_total = 0;
int count_bad = 0;

/* from geng.o and gentreeg.o */
int my_geng_main(int argc, char**argv);
int my_gentreeg_main(int argc, char**argv);

/*---------------------------------------------------------------------------*/

#define numberof(array)  (sizeof(array)/sizeof((array)[0]))

#define MAX(x,y)  ((x) > (y) ? (x) : (y))
#define MIN(x,y)  ((x) < (y) ? (x) : (y))

static const int vpar_singleton[] = {0,0};
static const int vpar_empty[] = {0};

/* Convert tree vpar[1..n] to sg using nautyextra_add_edge_sg(). */
void
test_vpar_to_sg_by_add (const int *vpar, int n, sparsegraph *sg)
{
  if (option_verbose)
    printf("test_vpar_to_sg_by_add() %p %d to %p\n", vpar, n, sg);
  sg->nv = sg->nde = 0;  /* emptygraph */
  nautyextra_ensure_vertex_sg(sg,n-1);
  int u;
  for (u = 1; u <= n; u++) {
    int v = vpar[u];
    if (v)
      nautyextra_add_edge_sg (sg, u-1, v-1);
  }
}

#define NAUTYEXTRA_SET1_LOOP(ss,n,body)                           \
  do {                                                            \
    assert (SETWORDSNEEDED(n) <= 1);                              \
    setword __NAUTYEXTRA_SET_LOOP__inc = (n ? BITT[n-1] : 0);     \
    (ss) = 0;                                                     \
    do {                                                          \
      body;                                                       \
    } while ((ss) += __NAUTYEXTRA_SET_LOOP__inc);                 \
  } while (0)


int
test_subtree_size_by_ancestors_vpar (const int *vpar, int n, int v)
{
  int weight = 0;
  int c;
  for (c = 1; c <= n; c++)
    weight += nautyextra_is_ancestor_vpar(vpar,c,v);
  return weight;
}

void
test_centroid_by_subtree_sizes_vpar (const int *vpar, int n,
                                     struct nautyextra_centroid_info *info)
{
  DEBUG(printf("test_centroid_by_subtree_sizes_vpar() ");
        nautyextra_print_vpar(vpar,n));

  DYNALLSTAT(int,subtree_sizes,subtree_sizes_sz);
  DYNALLOC1(int,subtree_sizes,subtree_sizes_sz,n+1,
            "test_centroid_by_subtree_sizes_vpar");
  DYNALLSTAT(int,weights,weights_sz);
  DYNALLOC1(int,weights,weights_sz,n+1,
            "test_centroid_by_subtree_sizes_vpar");

  nautyextra_subtree_sizes_array_vpar(vpar,n, subtree_sizes);
  DEBUG({int i;
          printf("subtree_sizes");
          for (i=1;i<=n;i++) printf(" %d",subtree_sizes[i]);
          printf("\n");});

  int v;
  for (v = 1; v <= n; v++)
    weights[v] = n - subtree_sizes[v];   /* everything above v */
  for (v = 1; v <= n; v++) {
    int p = vpar[v];
    if (p)
      weights[p] = MAX(weights[p], subtree_sizes[v]); /* child subtree sizes */
  }
  DEBUG({int i;
          printf("weights");
          for (i=1;i<=n;i++) printf(" %d",weights[i]);
          printf("\n");});

  int centroid_weight = n;
  int centroid_v1 = 0;
  int centroid_v2 = 0;
  for (v = 1; v <= n; v++) {
    if (centroid_weight > weights[v]) {   /* new low */
      centroid_weight = weights[v];
      centroid_v1 = v;
      centroid_v2 = v;
    } else if (centroid_weight == weights[v]) {     /* equal low */
      centroid_v2 = v;
    }
  }
  info->v1 = centroid_v1;
  info->v2 = centroid_v2;
  info->weight = centroid_weight;

#ifndef NDEBUG
  {
    int count = 0;
    for (v = 1; v <= n; v++)
      count += (weights[v]==centroid_weight);
    assert (n==0 ? count==0 : count==1 || count==2);
  }
#endif
}

static int
test_triangle_count_callback (const graph *g, int m, int n,
                              int a, int b, int c, void *data)
{
  (*(int*)data)++;
  DEBUG(printf("  triangle %d,%d,%d  (count %d)\n", a,b,c, *(int*)data));
  return 0;
}
/* same as nauty numtriangles() */
int
test_triangle_count (const graph *g, int m, int n)
{
  int count = 0;
  nautyextra_triangle_search (g,m,n, &test_triangle_count_callback, &count);
  return count;
}

struct test_triangle_list_by_search {
  int count;
  int *list;
  int limit;
};
static int
test_triangle_list_callback (const graph *g, int m, int n,
                             int a, int b, int c, void *data_parameter)
{
  struct test_triangle_list_by_search *data = data_parameter;
  if (data->count >= data->limit) abort();
  DEBUG(printf("  triangle %d,%d,%d  (count %d)\n", a,b,c, data->count));
  data->list[3*data->count]     = a;
  data->list[3*data->count + 1] = b;
  data->list[3*data->count + 2] = c;
  data->count++;
  return 0;
}
int
test_triangle_list_by_search (const graph *g, int m, int n,
                              int *list, int limit)
{
  DEBUG(printf("test_triangle_list_by_search()\n"));
  struct test_triangle_list_by_search data;
  data.count = 0;
  data.list = list;
  data.limit = limit;
  nautyextra_triangle_search (g,m,n, &test_triangle_list_callback, &data);
  return data.count;
}

int
test_triangle_list_by_abc (const graph *g, int m, int n,
                           int *list, int limit)
{
  DEBUG(printf("test_triangle_list_by_abc()\n"));
  int a,b,c, count = 0;
  for (a = 0; a < n; a++) {
    for (b = a+1; b < n; b++) {
      if (ISELEMENT(GRAPHROW(g,a,m),b)) {
        for (c = b+1; c < n; c++) {
          if (ISELEMENT(GRAPHROW(g,a,m),c)
              && ISELEMENT(GRAPHROW(g,b,m),c)) {
            DEBUG(printf("  triangle %d,%d,%d  (count %d)\n", a,b,c, count));
            if (count >= limit) abort();
            list[3*count]     = a;
            list[3*count + 1] = b;
            list[3*count + 2] = c;
            count++;
          }
        }
      }
    }
  }
  return count;
}

const int test_make_Beineke_maxn = 6;
int
test_make_Beineke (graph *g, int m, int num)
{
  assert (1 <= num && num <= 9);
  const struct {
    int n;
    struct edges {
      int u,v;
    } edges[11];
  } data[] = {
    {
      /* G1, claw */
      4,
      {{1,2},{1,3},{1,4}},
    },
    {
      /* G2 */
      5,
      {{1,2},{1,3},{1,4},
       {2,3},{2,5},
       {3,4},
       {4,5}},
    },
    {
      /* G3 */
      5,
      {{1,2},{1,3},{1,4},{1,5},
       {2,3},{2,4},{2,5},
       {3,4}, /* missing 3,5 */
       {4,5}},
    },
    {
      /* G4 */
      6,
      {{1,2},{1,3},{1,4},
       {2,3},{2,6},
       {3,4},
       {4,5}},
    },
    {
      /* G5 */
      6,
      {{1,2},{1,3},{1,4},
       {2,3},{2,4},{2,5},
       {3,4},
       {4,5},
       {5,6}},
    },
    {
      /* G6 */
      6,
      {{1,2},{1,3},{1,4},
       {2,3},{2,5},{2,4},{2,6},
       {3,4},
       {4,5},{4,6},
       {5,6}},
    },
    {
      /* G7 */
      6,
      {{1,2},{1,3},{1,4},
       {2,3},{2,6},
       {3,4},
       {4,5},
       {5,6}},
    },
    {
      /* G8 */
      6,
      {{1,2},{1,3},{1,4},{1,5},
       {2,3},
       {3,4},
       {4,5},{4,6},
       {5,6}},
    },
    {
      /* G9, wheel */
      6,
      {{1,2},{1,3},{1,4},{1,5},{1,6},  /* 1=middle */
       {2,3},{3,4},{4,5},{5,6},{6,2},},
    },
  };

  num--;
  EMPTYGRAPH(g, m, data[num].n);
  assert (data[num].n <= test_make_Beineke_maxn);
  int max_v = 0;
  int j;
  DEBUG(printf("  numberof edges %d\n", numberof(data[num].edges)));
  for (j = 0; j < numberof(data[num].edges); j++) {
    int u = data[num].edges[j].u;
    int v = data[num].edges[j].v;
    DEBUG(printf("  u=%d v=%d\n", u,v));
    assert ((u>0)==(v>0));
    if (! u) break;
    ADDONEEDGE(g, u-1,v-1, m);
    max_v = MAX(max_v, MAX(u,v));
  }
  assert (max_v == data[num].n);

  return max_v;

  /* for (j = 0; j < max_v; j++) */
  /* printf("%s\n", nautyextra_set_str(GRAPHROW(g,m,j),32)); */
}

int
test_is_linegraph_by_Beineke (const graph *g, int m, int n)
{
  const int sub_max_n = test_make_Beineke_maxn;
  const int sub_m = SETWORDSNEEDED(sub_max_n);
  graph sub_g[sub_max_n];
  int num;
  for (num = 1; num <= 9; num++) {
    int sub_n = test_make_Beineke(sub_g,sub_m, num);
    if (nautyextra_is_subgraph (sub_g,sub_m,sub_n, g,m,n, 1)) return FALSE;
  }
  return TRUE;
}

static int
INTERNAL_is_linegraph_callback (const graph *g, int m, int n,
                       int a, int b, int c, int d, void *data)
{
  /* return 1 to stop if found both odd (which means not line graph),
     0 to continue otherwise
  */
  return nautyextra_diamond_is_both_odd(g,m, a,b,c,d);
}
int
test_is_linegraph_by_claw_and_diamond (const graph *g, int m, int n)
{
  return nautyextra_is_clawfree (g,m,n)
    && ! nautyextra_diamond_search (g,m,n, 1,
                                    &INTERNAL_is_linegraph_callback, NULL);
}

int
test_height_num_components_by_search (const int *vpar, int n)
{
  DYNALLSTAT(int,heights,heights_sz);
  DYNALLOC1(int,heights,heights_sz,n+1,
            "test_height_num_components_by_search");
  nautyextra_subtree_heights_vpar (vpar,n, heights);

  int v, maximum_height = 0;
  for (v = 1; v <= n; v++)
    maximum_height = MAX(maximum_height, heights[v]);

  int num_roots = 0;
  for (v = 1; v <= n; v++)
    num_roots += (heights[v] == maximum_height);
  return num_roots;
}

setword
test_random_setword (int n)
{
  setword ss;
  GETKRAN(1<<n,ss);
  ss <<= WORDSIZE - n;
  assert ((ss & BITMASK(n)) == 0);
  return ss;
}

/*---------------------------------------------------------------------------*/
/* Independent Sets Helpers */

/* ss is a word where each bit is whether vertex v is in the set.
   Return 1 if this set is independent in sg, meaning any vertex pair u,v
   in the set has no edge u--v in sg.
   Return 0 if not.
*/
int
test_is_indset_sg1 (const sparsegraph *sg, setword ss)
{
  int n = sg->nv;
  int from,to;
  for (from=0; from<n; from++) {
    if (! (ss & (1<<from))) continue;
    for (to=from+1; to<n; to++) {
      if (! (ss & (1<<to))) continue;
      if (nautyextra_is_arc_sg(sg,from,to)) return FALSE;
    }
  }
  return TRUE;
}

NAUTYEXTRA_TYPE_ULL
test_indsets_count_by_all_sets (const graph *g, int m, int n)
{
  setword count = 0;
  setword ss;
  NAUTYEXTRA_SET1_LOOP (ss,n, count += nautyextra_is_indset(g,m,n,&ss));
  return count;
}

/* Set *info to the independence number of sg, and count of how many sets
   attain that independence number, by iterating all sets.
*/
struct test_indnum_count_sg_by_all_sets_info {
  int                 indnum;
  unsigned long long  count;
};
void
test_indnum_count_sg_by_all_sets (const sparsegraph *sg, struct test_indnum_count_sg_by_all_sets_info *info)
{
  setword count = 0;
  int indnum = 0;
  int n = sg->nv;
  setword ss;
  NAUTYEXTRA_SET1_LOOP
    (ss,n,
     if (nautyextra_is_indomset_sg(sg,&ss)) {
       int size = POPCOUNT(ss);
       if (size > indnum) {
         indnum = size;
         count = 0;
       }
       if (size == indnum) {
         count++;
       }
     });
  info->indnum = indnum;
  info->count  = count;
}

/* Return the independence number of a path of n vertices. */
int
test_path_indnum (int n)
{
  return (n<=1 ? n : (n+1)>>1); /* ceil(n/2) */
}
/* Return the number of independent sets attaining the independence number
   of a path of n vertices. */
unsigned long long
test_path_indnum_count (int n)
{
  /* 1 2 3 4 5 6 7 8
     *   *   *   *
     *   *   *     *
     *   *     *   *
     *     *   *   *
       *   *   *   *
  */
  return (n<=1 ? 1
          : n&1 ? 1
          : n==2 ? 2
          : (n>>1)+1);
}

/* Return the independence number of tree vpar[1..n].
 */
int
test_indnum_vpar_by_max (const int *vpar, int n)
{
  int i;
  struct data {
    int  num_children;
    int  indnum_with;
    int  indnum_without;
  };
#if MAXN
  struct data  data[MAXN+1];  /* 1 to n inclusive */
  assert(n <= MAXN);
#else
  DYNALLSTAT(struct data,data,data_sz);
  DYNALLOC1(struct data,data,data_sz,n+1,"nautyextra_indnum_vpar");
#endif
  assert (n >= 0);

  DEBUG(printf("nautyextra_indnum_vpar() n=%d", n);
        nautyextra_print_vpar(vpar,n));

  if (n==0) return 0;
  if (n==1) return 1;

  memset(data, '\0', (n+1)*sizeof(*data));
  for (i = 1; i <= n; i++) {
    data[vpar[i]].num_children++;
    data[i].indnum_with = 1;
    /* data[i].indnum_without = 0; */   /* already 0 */
  }

  int ret_with = 0;
  int ret_without = 0;
  for (i = 1; i <= n; i++) {
    int v = i;
    DEBUG(printf("v=%d num_children %d indnum %d,%d\n",
                 v, data[v].num_children,
                 data[v].indnum_with, data[v].indnum_without));
    if (data[v].num_children != 0) {
      /* either num_children=-1 for v already considered,
         or num_children>0 for v not yet a leaf */
      continue;
    }

    for (;;) {
      int c_with    = data[v].indnum_with;
      int c_without = data[v].indnum_without;
      DEBUG(printf("consider v=%d indnum %d,%d\n",
                   v, data[v].indnum_with, data[v].indnum_without));
      assert (data[v].num_children == 0);

      data[v].num_children = -1;
      v = vpar[v];     /* parent */
      if (! v) {
        DEBUG(printf("  no parent\n"));
        break;
      }

      ret_with = (data[v].indnum_with += c_without);

      int c_max_with_or_without = MAX(c_with, c_without);
      ret_without = (data[v].indnum_without += c_max_with_or_without);

      assert(data[v].num_children > 0);
      if (--data[v].num_children) {
        DEBUG(printf("  stop at non-leaf v=%d  indnum %d,%d\n",
                     v, ret_with, ret_without));
        break;
      }
    }
  }

  int ret = MAX(ret_with, ret_without);
  DEBUG(printf("  final %d,%d = %d\n",
               ret_with, ret_without, ret));
  return ret;
}

/*---------------------------------------------------------------------------*/
/* Independent Dominating Sets (Maximal Independent Sets) Helpers */

int
test_is_indomset_by_ind_and_dom_sg (const sparsegraph *sg, const set *ss)
{
  return nautyextra_is_indset_sg(sg,ss)
    && nautyextra_is_domset_sg(sg,ss);
}


/*---------------------------------------------------------------------------*/
/* Matching Helpers */

void
test_matchnum_and_count_vpar_by_sets (const int *vpar, int n,
                                    struct nautyextra_matchnum_and_count *info)
{
  if (n > WORDSIZE) {
    fprintf(stderr, "test_matchnum_and_count_vpar_by_sets() limit n <= %d, got n=%d\n",
            WORDSIZE, n);
    abort();
  }
  int matchnum = -1;
  setword count = 0;
  setword ss;
  NAUTYEXTRA_SET1_LOOP
    (ss,n,
     if (nautyextra_is_matching_vpar_set(vpar,n,&ss)) {
       int ss_size = POPCOUNT(ss);
       if (ss_size > matchnum) {
         matchnum = ss_size;
         count = 0;
       }
       if (ss_size == matchnum) {
         count++;
       }
     });
  info->matchnum = matchnum;
  info->count = count;
}

/*---------------------------------------------------------------------------*/
/* Dominating Sets Helpers */

/* ss is a word where each bit is whether vertex v is in the set.
   Return 1 if this set dominates in sg, meaning each vertex of sg is
   either in ss or adjancent to a vertex in ss.
   Return 0 if not.
*/
int
test_is_domset_sg1 (const sparsegraph *sg, setword ss)
{
  int n = sg->nv;
  int v;
  for (v=0; v<n; v++) {
    if (ISELEMENT(&ss,v)) continue;  /* v in ss */
    int degree = sg->d[v];
    int epos   = sg->v[v];
    int i;
    for (i=0; i<degree; i++) {
      int to = sg->e[epos+i];
      if (ISELEMENT(&ss,to)) goto NEXT_V;  /* neighbour "to" in ss */
    }
    return 0;
  NEXT_V: ;
  }
  return 1;
}

int
test_is_minimal_domset_sg1 (const sparsegraph *sg, setword ss)
{
  int n = sg->nv;
  assert (n >= 0);
#if MAXN
  int  dominated[MAXN];
  assert(n <= MAXN);
#else
  DYNALLSTAT(int,dominated,dominated_sz);
  DYNALLOC1(int,dominated,dominated_sz,n,"test_is_minimal_domset_sg1");
#endif

  memset(dominated,'\0', n*sizeof(dominated[0]));
  int v;
  for (v=0; v<n; v++) {       /* dominated[] count of v and neighbours */
    if (ISELEMENT(&ss,v)) {
      dominated[v]++;
      int degree = sg->d[v];
      int epos   = sg->v[v];
      int i;
      for (i=0; i<degree; i++) {
        int to = sg->e[epos+i];
        dominated[to]++;
      }
    }
  }
  for (v=0; v<n; v++) {
    if (!dominated[v]) return 0;  /* undominated */
    if (ISELEMENT(&ss,v) && dominated[v] >= 2) {
      int degree = sg->d[v];
      int epos   = sg->v[v];
      int i;
      for (i=0; i<degree; i++) {
        int to = sg->e[epos+i];
        if (dominated[to] < 2) goto NEXT_V;
      }
      return 0;  /* v and neighbours all >= 2 doms, v not minimal */
    }
  NEXT_V: ;
  }
  return 1;
}

/* Return the domination number of sg, found by iterating all sets and
   noting the smallest set which dominates.
*/
int
test_domnum_by_all_sets (const sparsegraph *sg)
{
  int n = sg->nv;
  int domnum = n;
  setword ss = 0;
  NAUTYEXTRA_SET1_LOOP(ss,n,
                       if (nautyextra_is_domset_sg(sg,&ss)) {
                         int num = POPCOUNT(ss);
                         if (num < domnum) domnum = num;
                       });
  return domnum;
}

/* Return the domination number of tree vpar[1..n], by considering min
   combinations accumulated into parent.

   (Not as good as Cockayne, Goodman, and Hedetniemi in nautyextra.c, but
   another approach to the same answer.)

   undom  0   1   1   1   2   2   2
    dom   1   1   1   2   2   2   3
    with  1   1   2   2   2   2   3
          1---2---3---4---5---6---7
          x   @

   with  = domnum with v present
   dom   = domnum with v dominated, either by v present or by some child
   undom = domnum with v not certain to be dominated, but could be
    undom <= dom <= with
   domextra = dom - undom >=0    increment from undom to dom
   withextra = with - dom >=0    increment from dom to with

   with      1 + sum c_undom
   undom         sum c_dom
   dom       children >=1 with, rest dom
             = sum child dom + min(child with-dom = withextra)
             = undom + min(child withextra)
   domextra  min(c_withextra)
   withextra = with - dom

                     *
                /    |     \
   3,3,1    *        *        *
           /|\      /|\      /|\
          * * *    * * *    * * *
   top 3,1,4
       6,2,7

   with           = 1 + sum c_possibly_undom
   possibly_undom = min(with, sum c_dom)
   dom            = min(with, sum c_dom + min(c_with-c_dom))
 */
int
test_domnum_vpar_by_min (const int *vpar, int n)
{
  int i;
  struct data {
    int  num_children;
    int  sum_c_undom;
    int  sum_c_dom;
    int  min_c_withextra;
  };
#if MAXN
  struct data  data[MAXN+1];
#else
  DYNALLSTAT(struct data,data,data_sz);

  DYNALLOC1(struct data,data,data_sz,n+1,"nautyextra_domnum_vpar");
#endif
  assert (n >= 0);

  DEBUG(printf("nautyextra_domnum_vpar() n=%d", n);
        nautyextra_print_vpar(vpar,n));

  if (n==0) return 0;
  if (n==1) return 1;

  memset(data, '\0', (n+1)*sizeof(*data));
  for (i = 1; i <= n; i++) {
    data[vpar[i]].num_children++;
    data[i].sum_c_undom     = 0;
    data[i].sum_c_dom       = 0;
    data[i].min_c_withextra = 1;
  }

  int ret_sum_c_undom     = 0;
  int ret_sum_c_dom       = 0;
  int ret_min_c_withextra = 0;
  for (i = 1; i <= n; i++) {
    int v = i;
    DEBUG(printf("v=%d num_children %d domnum %d,%d,%d\n",
                 v, data[v].num_children,
                 data[v].sum_c_undom,
                 data[v].sum_c_dom,
                 data[v].min_c_withextra));
    if (data[v].num_children != 0) {
      /* either num_children=-1 for v already considered,
         or num_children>0 for v not yet a leaf */
      continue;
    }

    for (;;) {
      int c_sum_c_undom     = data[v].sum_c_undom;
      int c_sum_c_dom       = data[v].sum_c_dom;
      int c_min_c_withextra = data[v].min_c_withextra;

      int c_with = 1 + c_sum_c_undom;
      int c_undom = MIN(c_with, c_sum_c_dom);
      int c_dom = MIN(c_with, c_undom + c_min_c_withextra);
      int c_withextra = c_with - c_dom;
      int c_domextra = c_dom - c_undom;

      DEBUG(printf("consider v=%d (parent %d) domnums %d,%d,%d  sums %d,%d min %d\n",
                   v, vpar[v], c_undom,c_dom,c_with,
                   c_sum_c_undom,c_sum_c_dom,c_min_c_withextra));
      assert (data[v].num_children == 0);
      assert (c_withextra >= 0);
      assert (c_domextra >= 0);
      assert (c_dom >= c_undom);
      assert (c_with >= c_dom);
      assert (c_with == c_dom + c_withextra);

      data[v].num_children = -1;
      v = vpar[v];     /* parent */
      if (! v) {
        DEBUG(printf("  no parent\n"));
        break;
      }

      ret_sum_c_dom   = (data[v].sum_c_dom += c_dom);
      ret_sum_c_undom = (data[v].sum_c_undom += c_undom);

      ret_min_c_withextra = data[v].min_c_withextra;
      data[v].min_c_withextra
        = ret_min_c_withextra = MIN(ret_min_c_withextra, c_withextra);


      assert(data[v].num_children > 0);
      if (--data[v].num_children) {
        DEBUG(printf("  stop at non-leaf v=%d   %d,%d,%d\n",
                     v, ret_sum_c_undom, ret_sum_c_dom, ret_min_c_withextra));
        break;
      }
    }
  }

  int ret_with = 1 + ret_sum_c_undom;
  int ret_undom = MIN(ret_with, ret_sum_c_dom);
  int ret_dom = MIN(ret_with, ret_undom + ret_min_c_withextra);

  int ret = ret_dom;
  DEBUG(printf("  final %d,%d,%d = %d\n",
               ret_undom,ret_dom,ret_with, ret));
  return ret;
}

/*---------------------------------------------------------------------------*/
/* Disjoint Dominating Sets Helpers */

void
test_disjoint_domnum_and_count_by_all_sets (const sparsegraph *sg,
                                            struct nautyextra_disjoint_domnum_and_count *info)
{
  int n = sg->nv;
  assert (n >= 0);
  if (n == 0) {
    info->disjoint_domnum = 0;
    info->count = 1;
    return;
  }

  int disjoint_domnum = n+1;
  unsigned long count = 0;

  int size = 0;
  setword ss1 = 0;
  setword ss2 = 0;
  setword lowbit = NAUTYEXTRA_SETWORD_HIGHBIT >> (n-1);
  for (;;) {
    setword bit = NAUTYEXTRA_SETWORD_HIGHBIT;
    for (;;) {
      if (ss1 & bit) {
        ss1 ^= bit;    /* 1 -> 2 */
        ss2 ^= bit;
        break;
      }
      if (ss2 & bit) {
        ss2 ^= bit;    /* 2 -> 0 and step down */
        size--;
        if (bit == lowbit) {
          if (count == 0) disjoint_domnum = -1;
          info->disjoint_domnum = disjoint_domnum;
          assert ((count&1) == 0);
          info->count = count >> 1;
          return;
        }
        bit >>= 1;
        continue;
      }
      ss1 ^= bit;      /* 0 -> 1 */
      size++;
      break;
    }

    assert ((ss1 & ss2) == 0);
    if (nautyextra_is_domset_sg(sg,&ss1)
        && nautyextra_is_domset_sg(sg,&ss2)) {
      if (size < disjoint_domnum) {
        disjoint_domnum = size;
        count = 1;
      } else if (size == disjoint_domnum) {
        count++;
      }
    }
  }
}


/*---------------------------------------------------------------------------*/

/* Set sg to the tree T1 from
   Marcin Krzywkowski, "Trees Having Many Minimal Dominating Sets".
   sg should be SG_INIT() already, but any existing contents are replaced.

           26
     24          25
   6 .. 11    18 .. 23
   0 .. 5     12 .. 17

   w is the number of 2-vertex arms off 24 and 25.
   k is the number of copies of T1, joined at a second from leaf like 6.

   FIXME: Bi-star 6,6 with extra vertex in each edge ...
*/
void
test_make_T1 (sparsegraph *sg, int w, int k)
{
  sg->nv = sg->nde = 0;
  int i;
  int v = 0;
  int join = 0;
  int to;

  while (k-- > 0) {
    for (i = 0; i < w; i++, v++)  nautyextra_add_edge_sg (sg,v,v+w);
    to = v + 3*w;
    for (i = 0; i < w; i++, v++)  nautyextra_add_edge_sg (sg,v,to);

    for (i = 0; i < w; i++, v++)  nautyextra_add_edge_sg (sg,v,v+w);
    to = v + w + 1;
    for (i = 0; i < w; i++, v++)  nautyextra_add_edge_sg (sg,v,to);

    nautyextra_add_edge_sg (sg,v,   v+2);
    nautyextra_add_edge_sg (sg,v+1, v+2);
    v += 3;

    if (join)
      nautyextra_add_edge_sg (sg,v-4,join);
    else
      join = v-4;
  }
  assert (sg->nde == (sg->nv - 1)*2);
}

int
test_vertex_eccentricity_by_path_length (const int *vpar, int n, int v)
{
  DEBUG(printf("test_vertex_eccentricity_by_path_length() v=%d\n", v));
  int eccentricity = 0;
  int to;
  for (to = 1; to <= n; to++) {
    int path_length = nautyextra_path_length_vpar(vpar,v,to);
    eccentricity = MAX(eccentricity,path_length);
    DEBUG(printf (" %d to %d path_length %d\n", v,to, path_length));
  }
  return eccentricity;
}
int
test_diameter_endpoints_by_eccentricity (const int *vpar, int n)
{
  int v;
  int diameter = nautyextra_diameter_vpar(vpar,n);
  int diameter_endpoints = 0;
  for (v = 1; v <= n; v++) {
    diameter_endpoints
      += (nautyextra_eccentricity_of_vertex_vpar(vpar,n,v) == diameter);
  }
  return diameter_endpoints;
}

/* Return the number of vertices which are in some diameter path,
   calculated by looking at all pairwise paths to find those which are
   diameters and flagging their vertices.
*/
int
test_diameter_vertices_by_paths (const int *vpar, int n)
{
  DEBUG(printf("test_diameter_vertices_by_paths() ");
        nautyextra_print_vpar(vpar,n));
  int m = SETWORDSNEEDED(n+1);
  setword seen[10];
  assert (m <= 10);
  if (m) EMPTYSET(seen,m);

  int degrees[100];
  assert (n < 100);
  nautyextra_degrees_array_vpar (vpar,n, degrees);

  int diameter = nautyextra_diameter_vpar(vpar,n);

  int path[100];
  assert (diameter <= 100);

  int u;
  for (u = 1; u <= n; u++) {
    if (degrees[u] > 1) continue;
    int v;
    for (v = u; v <= n; v++) {
      if (degrees[v] > 1) continue;
      int path_length = nautyextra_path_vertices_vpar(vpar, u,v, path);
      assert (path_length == nautyextra_path_length_vpar(vpar, u,v));
      assert (path_length >= 0);
      DEBUG({int i;
              printf ("%d->%d path_length %d  [",
                      u,v, path_length);
              for(i=1;i<=path_length;i++) printf("%d ",path[i]);
              printf("]\n");});
      if (path_length != diameter) continue;
      int i;
      for (i=0; i<=path_length; i++) {
        DEBUG(printf(" %d\n", path[i]));
        assert (1 <= path[i] && path[i] <= n);
        ADDELEMENT(seen,path[i]);
      }
    }
  }
  DEBUG(printf(" %s\n", nautyextra_set_str(seen,n+1)));
  return setsize(seen,m);
}


/*---------------------------------------------------------------------------*/

const int max_bad = 1;
SG_DECL(sg);
SG_DECL(sg2);
SG_DECL(canon_sg);
SG_DECL(canon_sg2);
int count_trees[64];

void
test_tree (const int *vpar, int n)
{
  if (option_verbose) {
    printf("tree vpar=%p  n=%d\n", vpar,n);
    nautyextra_print_vpar(vpar,n);
  }
  nautyextra_validate_vpar(vpar,n);

  nautyextra_vpar_to_sg(vpar,n,&sg,0); /* undirected */
  if (option_verbose) {
    nautyextra_print_sg(&sg);
  }
  assert(sg.nv   == n);
  assert(sg.nde  == (n>=1 ? 2*(n-1) : 0));
  /* assert(sg.elen == sg.nde); */  /* no holes in ee */
  nautyextra_validate_sg(&sg);
  assert(nautyextra_is_multiedged_sg(&sg) == FALSE);
  assert(nautyextra_loopcount_sg(&sg) == 0);

  DYNALLSTAT(graph,g,g_sz);
  int m = SETWORDSNEEDED(n);
  DYNALLOC2(graph,g,g_sz,m,n,"malloc");
  nautyextra_vpar_to_graph(vpar,g,m,n, 0); /* undirected */
  assert(loopcount(g,m,n) == 0);

  count_total++;
  count_trees[n]++;
  /* if (nautyextra_is_connected_vpar(vpar,n)) */
  /*   count_connected_trees[n]++; */

  {
    /* nautyextra_perfect_domnum_vpar()
       vs nautyextra_perfect_domnum_and_count_sg() */

    struct nautyextra_perfect_domnum_and_count  info_sg;
    nautyextra_perfect_domnum_and_count_sg(&sg, &info_sg);

    int perfect_domnum_by_vpar = nautyextra_perfect_domnum_vpar(vpar,n);
    if (perfect_domnum_by_vpar != info_sg.perfect_domnum) {
      printf("n=%d nautyextra_perfect_domnum_vpar() different\n", n);
      printf("  perfect_domnum_by_vpar   %d\n", perfect_domnum_by_vpar);
      printf("  info_sg.perfect_domnum  %d\n", info_sg.perfect_domnum);
      if (++count_bad >= max_bad) assert(0);
    }
  }
  {
    /* nautyextra_semitotdomnum_vpar()
       vs nautyextra_semitotdomnum_and_count_sg() */

    struct nautyextra_semitotdomnum_and_count  info_sg;
    nautyextra_semitotdomnum_and_count_sg(&sg, &info_sg);

    int semitotdomnum_by_vpar = nautyextra_semitotdomnum_vpar(vpar,n);
    if (semitotdomnum_by_vpar != info_sg.semitotdomnum) {
      printf("n=%d nautyextra_semitotdomnum_vpar() different\n", n);
      printf("  semitotdomnum_by_vpar   %d\n", semitotdomnum_by_vpar);
      printf("  info_sg.semitotdomnum  %d\n", info_sg.semitotdomnum);
      if (++count_bad >= max_bad) assert(0);
    }
  }

  {
    /* nautyextra_subtree_sizes_array_vpar() vs nautyextra_subtree_size_vpar()
     */
    int weights[tree_max_vertices+1];
    nautyextra_subtree_sizes_array_vpar(vpar,n, weights);
    int rep;
    for (rep = 1; rep <= MIN(n,5); rep++) {
      int v = KRAN(n)+1;
      assert (1 <= v && v <= n);
      int by_one = nautyextra_subtree_size_vpar(vpar,n, v);
      int by_all = weights[v];
      if (by_one != by_all) {
        printf("n=%d subtree_sizes\n", n);
        printf("  by_one  %d\n", by_one);
        printf("  by_all  %d\n", by_all);
        if (++count_bad >= max_bad) assert(0);
      }
    }
  }

  {
    /* nautyextra_height_num_components_vpar() */
    int got  = nautyextra_height_num_components_vpar(vpar,n);
    int want = test_height_num_components_by_search(vpar,n);
    assert (got == want);
  }

  if (n >= 1) {
    /* nautyextra_relabel_swap_vpar() */
    int u, v;
    GETKRAN(n,u); u++;   /* 1 to n */
    GETKRAN(n,v); v++;   /* 1 to n */

    int  swapped_vpar[tree_max_vertices+1];
    memcpy(swapped_vpar,vpar,(n+1)*sizeof(vpar[0]));
    nautyextra_relabel_swap_vpar (swapped_vpar,n,u,v);

    int  got_perm[tree_max_vertices+1];
    if (! nautyextra_is_isomorphic_vpar (vpar,swapped_vpar,n, 1, got_perm)) {
      printf("nautyextra_relabel_swap_vpar() n=%d u=%d,v=%d not isomorphic\n",
             n, u, v);
      if (++count_bad >= max_bad) assert(0);

    } else {
      int  vpar2_again[tree_max_vertices+1];
      nautyextra_relabel_perm_vpar (vpar, n, got_perm, vpar2_again);
      if (! nautyextra_aresame_vpar(vpar2_again, swapped_vpar, n)) {
        printf("nautyextra_relabel_swap_vpar() n=%d u=%d,v=%d perm does not restore\n",
               n, u, v);
        if (++count_bad >= max_bad) assert(0);
      }
    }

    int  i, perm[tree_max_vertices+1];
    for (i = 0; i <= n; i++)
      perm[i] = (i==u ? v : i==v ? u : i);
    int  permed_vpar[tree_max_vertices+1];
    nautyextra_relabel_perm_vpar (vpar,n, perm, permed_vpar);
    if (memcmp(swapped_vpar+1, permed_vpar+1, n*sizeof(swapped_vpar[0])) != 0) {
      printf("nautyextra_relabel_swap_vpar() n=%d u=%d,v=%d different to nautyextra_relabel_perm_vpar()\n",
             n, u, v);
      printf("vpar         "); nautyextra_print_vpar(vpar,n);
      printf("swapped_vpar "); nautyextra_print_vpar(swapped_vpar,n);
      printf("permed_vpar  "); nautyextra_print_vpar(permed_vpar,n);
      if (++count_bad >= max_bad) assert(0);
    }
  }
  {
    /* nautyextra_limb_length_vpar() vs reroot */

    int  degrees[tree_max_vertices+1];
    nautyextra_degrees_array_vpar (vpar,n, degrees);
    int  rerooted_vpar[tree_max_vertices+1];
    memcpy(rerooted_vpar,vpar,(n+1)*sizeof(vpar[0]));
    int v, rerooted_to = 0, max_degree = 0;
    for (v = 1; v <= n; v++) {
      if (degrees[v] != 2 && degrees[v] > max_degree) {
        max_degree = degrees[v];
        nautyextra_reroot_vpar(rerooted_vpar,n, v);
        rerooted_to = v;
      }
    }
    nautyextra_validate_vpar(rerooted_vpar,n);
    int by_reroot = nautyextra_limb_height_vpar(rerooted_vpar,n);

    int by_func = nautyextra_limb_length_vpar(vpar,n);
    if (by_func != by_reroot) {
      printf("n=%d nautyextra_limb_length_vpar() different\n", n);
      printf("  by_func   %d\n", by_func);
      printf("  by_reroot %d  root %d degree %d\n",
             by_reroot, rerooted_to, degrees[rerooted_to]);
      nautyextra_view_vpar(rerooted_vpar,n);
      nautyextra_view_synchronous = 1;
      nautyextra_view_vpar(vpar,n);
      if (++count_bad >= max_bad) assert(0);
    }
  }

  {
    /* nautyextra_totdomnum_vpar() and
       nautyextra_totdomnum_and_count_vpar()
       vs nautyextra_totdomnum_and_count_sg() */

    struct nautyextra_totdomnum_and_count  info_sg;
    nautyextra_totdomnum_and_count_sg(&sg, &info_sg);

    int totdomnum_by_vpar = nautyextra_totdomnum_vpar(vpar,n);
    if (totdomnum_by_vpar != info_sg.totdomnum) {
      printf("n=%d nautyextra_totdomnum_vpar() different\n", n);
      printf("  totdomnum_by_vpar   %d\n", totdomnum_by_vpar);
      printf("  info_sg.totdomnum  %d\n", info_sg.totdomnum);
      if (++count_bad >= max_bad) assert(0);
    }

    struct nautyextra_totdomnum_and_count  info_vpar;
    nautyextra_totdomnum_and_count_vpar(vpar,n, &info_vpar);
    if (info_vpar.totdomnum != info_sg.totdomnum
        || info_vpar.count != info_sg.count) {
      printf("n=%d nautyextra_totdomnum_and_count_vpar() different\n", n);
      printf("  info_vpar  %d count %Lu\n",
             info_vpar.totdomnum, info_vpar.count);
      printf("  info_sg    %d count %Lu\n",
             info_sg.totdomnum, info_sg.count);
      nautyextra_view_vpar(vpar,n);
      if (++count_bad >= max_bad) assert(0);
    }
  }

  {
    /* nautyextra_indomnum_vpar()
       nautyextra_indomnum_and_count_vpar()
       vs nautyextra_indomnum_and_count_sg() */

    struct nautyextra_indomnum_and_count  info_sg;
    nautyextra_indomnum_and_count_sg(&sg, &info_sg);

    struct nautyextra_indomnum_and_count  info_vpar;
    nautyextra_indomnum_and_count_vpar(vpar,n, &info_vpar);
    if (info_vpar.indomnum != info_sg.indomnum
        || info_vpar.count != info_sg.count) {
      printf("n=%d nautyextra_indomnum_and_count_vpar() different\n", n);
      printf("  info_vpar  %d count %Lu\n",
             info_vpar.indomnum, info_vpar.count);
      printf("  info_sg    %d count %Lu\n",
             info_sg.indomnum, info_sg.count);
      nautyextra_view_vpar(vpar,n);
      if (++count_bad >= max_bad) assert(0);
    }

    int indomnum_by_vpar = nautyextra_indomnum_vpar(vpar,n);
    if (indomnum_by_vpar != info_sg.indomnum) {
      printf("n=%d nautyextra_indomnum_vpar() different\n", n);
      printf("  indomnum_by_vpar  %d\n", indomnum_by_vpar);
      printf("  info_sg.indomnum  %d\n", info_sg.indomnum);
      nautyextra_view_vpar(vpar,n);
      if (++count_bad >= max_bad) assert(0);
    }
  }

  {
    if (option_verbose) printf ("depths\n");

    /* nautyextra_depths_vpar() vs nautyextra_depth_of_vertex_vpar(),
       and nautyextra_height_vpar() */
    int depths[tree_max_vertices+1];
    int v;
    nautyextra_depths_vpar(vpar,n, depths);
    int diff = 0;
    for (v = 1; v <= n; v++) {
      int by_one = nautyextra_depth_of_vertex_vpar(vpar, v);
      if (depths[v] != by_one) {
        printf("WRONG n=%d v=%d depths\n", n, v);
        printf("  by_one    %d\n", by_one);
        printf("  depths[]  %d\n", depths[v]);
        diff = 1;
      }
    }
    if (diff) {
      nautyextra_print_vpar(vpar,n);
      if (++count_bad >= max_bad) assert(0);
    }

    int height_by_depths = -1;
    int height_by_depths_v = 0;
    for (v = 1; v <= n; v++) {
      if (depths[v] > height_by_depths) {
        height_by_depths = depths[v];
        height_by_depths_v = v;
      }
    }
    {
      int got_height = nautyextra_height_vpar(vpar,n);
      if (got_height != height_by_depths) {
        printf("n=%d got_height %d height_by_depths %d\n",
               n, got_height, height_by_depths);
        if (++count_bad >= max_bad) assert(0);
      }
    }

    int height2_by_depths = -1;
    for (v = 1; v <= n; v++) {
      if (v != height_by_depths_v)
        height2_by_depths = MAX(height2_by_depths, depths[v]);
    }
    {
      int got_2height = nautyextra_2height_vpar(vpar,n);
      if (got_2height != height2_by_depths) {
        printf("n=%d got_height2 %d height2_by_depths %d\n",
               n, got_2height, height2_by_depths);
        if (++count_bad >= max_bad) assert(0);
      }
    }
  }

  {
    /* nautyextra_eccentricity_of_vertex() vs rerooted height vs path length */
    int  rerooted_vpar[tree_max_vertices+1];
    memcpy(rerooted_vpar,vpar,(n+1)*sizeof(vpar[0]));
    int v;
    int diff = 0;
    for (v = 1; v <= n; v++) {
      int eccentricity = nautyextra_eccentricity_of_vertex_vpar(vpar,n, v);
      {
        nautyextra_reroot_vpar(rerooted_vpar,n, v);
        int height = nautyextra_height_vpar(rerooted_vpar,n);
        if (eccentricity != height) {
          printf("WRONG n=%d v=%d\n", n, v);
          printf("  reroot height  %d\n", height);
          printf("  eccentricity   %d\n", eccentricity);
          diff++;
          break;
        }
      }
      {
        int by_path_length = test_vertex_eccentricity_by_path_length(vpar,n,v);
        if (eccentricity != by_path_length) {
          printf("WRONG n=%d v=%d\n", n, v);
          printf("  by_path_length %d\n", by_path_length);
          printf("  eccentricity   %d\n", eccentricity);
          diff++;
          break;
        }
      }
    }
    if (diff) {
      nautyextra_view_vpar(vpar,n);
      nautyextra_print_vpar(vpar,n);
      if (++count_bad >= max_bad) assert(0);
    }
  }

  {
    /* nautyextra_eccentricities_vpar() vs individual */
    int  eccs[tree_max_vertices+1];
    nautyextra_eccentricities_vpar(vpar,n, eccs);
    int diff = 0;
    int v;
    for (v = 1; v <= n; v++) {
      int ecc = nautyextra_eccentricity_of_vertex_vpar(vpar,n, v);
      if (eccs[v] != ecc) {
        printf("WRONG n=%d v=%d\n", n, v);
        printf("  eccs[v]  %d\n", eccs[v]);
        printf("  ecc      %d\n", ecc);
        diff++;
        break;
      }
    }
    if (diff) {
      nautyextra_view_vpar(vpar,n);
      nautyextra_print_vpar(vpar,n);
      if (++count_bad >= max_bad) assert(0);
    }
  }

  {
    /* nautyextra_centroid_info_vpar() vs test_centroid_by_subtree_sizes_vpar()
     */
    struct nautyextra_centroid_info  info_func, info_test;
    nautyextra_centroid_info_vpar(vpar,n, &info_func);
    test_centroid_by_subtree_sizes_vpar(vpar,n, &info_test);
    if (! (0 <= info_func.v1 && info_func.v1 <= n
           && 0 <= info_func.v2 && info_func.v2 <= n
           && info_func.v1 <= info_func.v2
           && info_func.weight >= 0
           && info_func.weight <= (n==0 ? 0 : n-1)
           && info_func.v1 == info_test.v1
           && info_func.v2 == info_test.v2
           && info_func.weight == info_test.weight)) {
      printf("n=%d centroid\n", n);
      printf("  by_func  v=%d,%d weight %d\n",
             info_func.v1, info_func.v2, info_func.weight);
      printf("  by_test  v=%d,%d weight %d\n",
             info_test.v1, info_test.v2, info_test.weight);
      if (++count_bad >= max_bad) assert(0);
    }
  }

  {
    /* nautyextra_subtree_size_vpar() vs test_subtree_size_by_ancestors_vpar()
     */
    int rep;
    for (rep = 1; rep <= MIN(n,5); rep++) {
      int v = KRAN(n)+1;
      assert (1 <= v && v <= n);
      int by_func = nautyextra_subtree_size_vpar(vpar,n, v);
      int by_test = test_subtree_size_by_ancestors_vpar(vpar,n, v);
      if (by_func != by_test) {
        printf("n=%d subtree_size\n", n);
        printf("  by_func  %d\n", by_func);
        printf("  by_test  %d\n", by_test);
        if (++count_bad >= max_bad) assert(0);
      }
    }
  }

  if (n <= 10) {
    /* nautyextra_claw_count_vpar() vs nautyextra_claw_count()
       vs nautyextra_is_clawfree_vpar()
    */
    int by_vpar  = nautyextra_claw_count_vpar(vpar,n);
    int by_graph = nautyextra_claw_count(g,m,n);
    if (by_vpar != by_graph) {
      printf("n=%d claw_count\n", n);
      printf("  by_vpar   %d\n", by_vpar);
      printf("  by_graph  %d\n", by_graph);
      /* nautyextra_view(g,m,n); */
      if (++count_bad >= max_bad) assert(0);
    }

    int is_clawfree_by_count = (by_vpar==0);
    int is_clawfree = nautyextra_is_clawfree_vpar(vpar,n);
    if (is_clawfree_by_count != is_clawfree) {
      printf("n=%d nautyextra_is_clawfree_vpar", n);
      if (++count_bad >= max_bad) assert(0);
    }

    const int claw_n = 4;
    const int claw_m = 1;
    graph claw_g[claw_m];
    nautyextra_make_star(claw_g,claw_m,claw_n, 0);
    int is_clawfree_by_subgraph
      = ! nautyextra_is_subgraph(claw_g,claw_m,claw_n, g,m,n, 1);
    if (is_clawfree_by_subgraph != is_clawfree) {
      printf("n=%d is_clawfree by subgraph\n", n);
      printf("  got  %d\n", is_clawfree_by_subgraph);
      printf("  want %d\n", is_clawfree);
      if (++count_bad >= max_bad) assert(0);
    }
  }

  {
    /* any tree extended by 1 extra vertex at each existing gives 2^n many
       minimal dominating sets
    */
    const int max_n = 16;
    if (n <= max_n) {
      int vpar_ext[2*max_n + 1];
      int n_ext = 2*n;
      memcpy(vpar_ext, vpar, sizeof(vpar[0])*(n+1));
      nautyextra_join_paths_vpar (vpar_ext, n, 1);
      unsigned long long  by_vpar
        = nautyextra_minimal_domsets_count_vpar(vpar_ext,n_ext);
      unsigned long long  by_formula = 1ULL << n;
      if (by_vpar != by_formula) {
        printf("n=%d minimal domsets of tree extended\n", n);
        printf("  by_vpar     %Lu\n", by_vpar);
        printf("  by_formula  %Lu\n", by_formula);
        nautyextra_view_sg(&sg);
        if (++count_bad >= max_bad) assert(0);
      }
    }
  }

  if (n <= 8) {
    /* nautyextra_disjoint_domnum_vpar()
       nautyextra_disjoint_domnum_and_count_vpar()
     */
    struct nautyextra_disjoint_domnum_and_count info_by_test;
    test_disjoint_domnum_and_count_by_all_sets(&sg,&info_by_test);

    int by_vpar = nautyextra_disjoint_domnum_vpar(vpar,n);
    if (by_vpar != info_by_test.disjoint_domnum) {
      printf("n=%d disjoint domnum\n", n);
      printf("  by_vpar  %d\n", by_vpar);
      printf("  by_test  %d\n", info_by_test.disjoint_domnum);
      nautyextra_view_sg(&sg);
      if (++count_bad >= max_bad) assert(0);
    }

    struct nautyextra_disjoint_domnum_and_count info_by_vpar;
    nautyextra_disjoint_domnum_and_count_vpar(vpar,n,&info_by_vpar);
    if (info_by_vpar.disjoint_domnum != info_by_test.disjoint_domnum
        || info_by_vpar.count        != info_by_test.count) {
      printf("n=%d disjoint domnum\n", n);
      printf("  info_by_vpar  %d %Lu\n",
             info_by_vpar.disjoint_domnum, info_by_vpar.count);
      printf("  info_by_test  %d %Lu\n",
             info_by_test.disjoint_domnum, info_by_test.count);
      nautyextra_view_sg(&sg);
      if (++count_bad >= max_bad) assert(0);
    }
  }

  {
    int matchnum_by_vpar = nautyextra_matchnum_vpar(vpar,n);

    struct nautyextra_matchnum_and_count  info_vpar;
    nautyextra_matchnum_and_count_vpar(vpar,n, &info_vpar);

    if (matchnum_by_vpar != info_vpar.matchnum) {
      printf("n=%d nautyextra_matchnum_vpar() different from nautyextra_matchnum_and_count_vpar()\n", n);
      if (++count_bad >= max_bad) assert(0);
    }

    if (n < 12) {
      /* vs test_matchnum_and_count_vpar_by_sets() */
      struct nautyextra_matchnum_and_count  info_test;
      test_matchnum_and_count_vpar_by_sets(vpar,n,&info_test);
      if (info_vpar.matchnum != info_test.matchnum
          || info_vpar.count != info_test.count) {
        printf("n=%d nautyextra_matchnum_vpar() different from sets\n", n);
        printf("  by_vpar  %d count %Lu\n", info_vpar.matchnum,info_vpar.count);
        printf("  by sets  %d count %Lu\n", info_test.matchnum,info_test.count);
        nautyextra_print_vpar(vpar,n);
        nautyextra_view_sg(&sg);
        if (++count_bad >= max_bad) assert(0);
      }
    }
  }

  if (n < 12) {
    /* nautyextra_minimal_domsets_count_vpar()
       vs nautyextra_minimal_domsets_count_sg() */
    unsigned long long by_vpar = nautyextra_minimal_domsets_count_vpar(vpar,n);
    unsigned long long by_sg   = nautyextra_minimal_domsets_count_sg(&sg);
    if (by_vpar != by_sg) {
      printf("n=%d nautyextra_minimal_domsets_count_vpar() different\n", n);
      printf("  by_vpar  %Lu\n", by_vpar);
      printf("  by_sg    %Lu\n", by_sg);
      printf("vpar "); nautyextra_print_vpar(vpar,n);
      printf("%s", sgtog6(&sg));
      /* nautyextra_view_sg(&sg); */
      if (++count_bad >= max_bad) assert(0);
    }
  }

  if (n < 12) {
    /* nautyextra_domsets_count_vpar() vs nautyextra_domsets_count_sg() */

    unsigned long long by_vpar = nautyextra_domsets_count_vpar(vpar,n);
    unsigned long long by_sg   = nautyextra_domsets_count_sg(&sg);
    if (by_vpar != by_sg) {
      printf("n=%d nautyextra_domsets_count_vpar() different\n", n);
      printf("  by_vpar  %Lu\n", by_vpar);
      printf("  by_sg    %Lu\n", by_sg);
      /* nautyextra_view_sg(&sg); */
      if (++count_bad >= max_bad) assert(0);
    }
  }

  {
    /* nautyextra_domnum_vpar() vs test_domnum_vpar_by_min() */

    int by_vpar = nautyextra_domnum_vpar(vpar,n);
    int by_min  = test_domnum_vpar_by_min(vpar,n);
    if (by_vpar != by_min) {
      printf("n=%d nautyextra_domnum_vpar() different\n", n);
      printf("  by_vpar  %d\n", by_vpar);
      printf("  by_min   %d\n", by_min);
      if (++count_bad >= max_bad) assert(0);
    }
  }
  if (n < 12) {
    /* nautyextra_domnum_vpar() vs
       nautyextra_domnum_and_count_vpar() vs
       nautyextra_domnum_and_count_sg()
    */
    int domnum_by_vpar = nautyextra_domnum_vpar(vpar,n);

    struct nautyextra_domnum_and_count  info_sg;
    nautyextra_domnum_and_count_sg(&sg,&info_sg);
    if (domnum_by_vpar != info_sg.domnum) {
      printf("n=%d nautyextra_domnum_vpar() and nautyextra_domnum_and_count_sg() different\n", n);
      printf("  by_vpar  %d\n", domnum_by_vpar);
      printf("  by_sg    %d\n", info_sg.domnum);
      if (++count_bad >= max_bad) assert(0);
    }

    struct nautyextra_domnum_and_count  info_vpar;
    nautyextra_domnum_and_count_vpar(vpar,n,&info_vpar);
    if (info_vpar.domnum != info_sg.domnum || info_vpar.count != info_sg.count) {
      printf("n=%d nautyextra_domnum_and_count_vpar() and nautyextra_domnum_and_count_sg() different\n", n);
      printf("  by_vpar  %d count %Lu\n", info_vpar.domnum, info_vpar.count);
      printf("  by_sg    %d count %Lu\n", info_sg.domnum, info_sg.count);
      nautyextra_print_vpar(vpar,n);
      nautyextra_view_vpar(vpar,n);
      if (++count_bad >= max_bad) assert(0);
    }
  }

  if (n < 12) {
    /* nautyextra_domnum_vpar() vs test_domnum_by_all_sets() search */

    int by_vpar = nautyextra_domnum_vpar(vpar,n);
    int by_sg   = test_domnum_by_all_sets(&sg);
    if (by_vpar != by_sg) {
      printf("n=%d nautyextra_domnum_vpar() different\n", n);
      printf("  by_vpar  %d\n", by_vpar);
      printf("  by_sg    %d\n", by_sg);
      if (++count_bad >= max_bad) assert(0);
    }
  }

  if (n < 12) {
    /* nautyextra_indsets_count_vpar()
       vs nautyextra_indsets_count_sg()
       vs test_indsets_count_by_all_sets() */

    unsigned long long by_vpar = nautyextra_indsets_count_vpar(vpar,n);
    unsigned long long by_sg   = nautyextra_indsets_count_sg(&sg);
    unsigned long long by_all  = test_indsets_count_by_all_sets(g,m,n);
    if (by_vpar != by_sg || by_vpar != by_all) {
      printf("n=%d nautyextra_indsets_count_vpar() different\n", n);
      printf("  by_vpar  %Lu\n", by_vpar);
      printf("  by_sg    %Lu\n", by_sg);
      printf("  by_all   %Lu\n", by_all);
      if (++count_bad >= max_bad) assert(0);
    }
  }

  if (n < 12) {
    /* nautyextra_indomsets_count_vpar() vs simple count on sg */

    unsigned long long by_vpar = nautyextra_indomsets_count_vpar(vpar,n);
    unsigned long long by_sg   = nautyextra_indomsets_count_sg(&sg);
    if (by_vpar != by_sg) {
      printf("n=%d nautyextra_indomsets_count_vpar() different\n", n);
      printf("  by_vpar  %Lu\n", by_vpar);
      printf("  by_sg    %Lu\n", by_sg);
      if (++count_bad >= max_bad) assert(0);
    }

    unsigned long long want_max = nautyextra_indomsets_count_tree_max (n);
    assert (by_vpar <= want_max);
  }

  struct test_indnum_count_sg_by_all_sets_info info_by_all_sets;
  test_indnum_count_sg_by_all_sets(&sg,&info_by_all_sets);
  {
    /* nautyextra_indnum_vpar() and nautyextra_indnum_and_count_vpar() */

    int indnum_by_vpar = nautyextra_indnum_vpar(vpar,n);
    assert (indnum_by_vpar >= 0);
    assert (indnum_by_vpar <= (n<=1 ? n : n-1));

    struct nautyextra_indnum_and_count  info_by_vpar;
    nautyextra_indnum_and_count_vpar(vpar,n, &info_by_vpar);
    assert (info_by_vpar.indnum == indnum_by_vpar);
    assert (info_by_vpar.count > 0);

    int indnum_by_max = test_indnum_vpar_by_max(vpar,n);
    assert (indnum_by_max >= 0);
    assert (indnum_by_max <= (n<=1 ? n : n-1));

    if (indnum_by_vpar != info_by_all_sets.indnum) {
      printf("n=%d nautyextra_indnum_vpar() different\n", n);
      printf("  indnum_by_vpar  %d\n", indnum_by_vpar);
      printf("  indnum by sg    %d\n", info_by_all_sets.indnum);
      if (++count_bad >= max_bad) assert(0);
    }
    if (info_by_vpar.count != info_by_all_sets.count) {
      printf("n=%d nautyextra_indnum_count_vpar() different\n", n);
      printf("  count_by_vpar  %Lu\n", info_by_vpar.count);
      printf("  count by sg    %Lu\n", info_by_all_sets.count);
      if (++count_bad >= max_bad) assert(0);
    }
    if (indnum_by_max != info_by_all_sets.indnum) {
      printf("n=%d test_indnum_vpar_by_max() different\n", n);
      printf("  indnum_by_max  %d\n", indnum_by_max);
      printf("  indnum by sg    %d\n", info_by_all_sets.indnum);
      if (++count_bad >= max_bad) assert(0);
    }

    if (n >= 4) {
      /* Jou and Lin, tree n>=4 without duplicated leaf has
         indnum <= floor((2n-1)/3)
      */
      int indnum_limit = (2*n-1)/3;
      if (! nautyextra_any_duplicated_leaf_vpar(vpar,n)) {
        assert (indnum_by_vpar <= indnum_limit);
      }
    }
  }
  {
    /* nautyextra_indnum_and_count() */
    struct nautyextra_indnum_and_count  info;
    nautyextra_indnum_and_count(g,m,n,&info);
    assert (info.indnum >= 0);
    assert (info.indnum <= (n<=1 ? n : n-1));

    if (info.indnum != info_by_all_sets.indnum) {
      printf("n=%d nautyextra_indnum_vpar() different\n", n);
      printf("  info.indnum     %d\n", info.indnum);
      printf("  indnum by sg    %d\n", info_by_all_sets.indnum);
      if (++count_bad >= max_bad) assert(0);
    }
    if (info.count != info_by_all_sets.count) {
      printf("n=%d nautyextra_indnum_count_vpar() different\n", n);
      printf("  info.count          %Lu\n", info.count);
      printf("  indnum_count by sg  %Lu\n", info_by_all_sets.count);
      if (++count_bad >= max_bad) assert(0);
    }
  }

  {
    /* nautyextra_sg_to_vpar() inverse of nautyextra_vpar_to_sg() */
    int root;
    for (root = 0; root < n; root++) {
      int vpar2[tree_max_vertices+1];
      nautyextra_sg_to_vpar(&sg, vpar2, root);
      if (! nautyextra_is_isomorphic_vpar2 (vpar,n, vpar2,n, 0)) {
        printf("n=%d vpar to g and sg not isomorphic\n", n);
        if (++count_bad >= max_bad) assert(0);
      }
    }
  }

  {
    /* nautyextra_vpar_to_graph() == nautyextra_vpar_to_sg() */
    if (n==0)
      sg.nv = sg.nde = 0;  /* nauty_to_sg() doesn't take n=0 */
    else
      nauty_to_sg (g,&sg,m,n);
    if (! nautyextra_is_isomorphic_gsg (g,&sg,m,n)) {
      printf("n=%d vpar to g and sg not isomorphic\n", n);
      if (++count_bad >= max_bad) assert(0);
    }
  }

  {
    /* nautyextra_is_cyclic() == 0 on tree */

    if (! (nautyextra_is_cyclic(g,m,n) == FALSE)) {
      printf("n=%d g tree not cyclic\n", n);
      if (++count_bad >= max_bad) assert(0);
    }
  }

  {
    /* nautyextra_vpar_to_sg() == build sg by adding edges */

    test_vpar_to_sg_by_add (vpar,n,&sg2);
    if (option_verbose) {
      printf("sg2\n");
      nautyextra_print_sg(&sg2);
    }
    nautyextra_validate_sg(&sg2);

    if (! (nautyextra_is_isomorphic_sg (&sg,&sg2) == TRUE)) {
      printf("n=%d vpar to sg and by add edges not isomorphic\n", n);
      nautyextra_print_sg(&sg);
      printf("sg2\n");
      nautyextra_print_sg(&sg2);
      if (++count_bad >= max_bad) assert(0);
    }
  }

  int g_radius, g_diameter;
  diamstats(g, m,n, &g_radius, &g_diameter);

  {
    /* nautyextra_Wiener_index() == nautyextra_Wiener_index_sg() */
    int terminal;
    for (terminal = 0; terminal <= 0; terminal++) {
      struct nautyextra_Wiener_info  info;
      nautyextra_Wiener_index(g,m,n, terminal, &info);

      struct nautyextra_Wiener_info info_sg;
      nautyextra_Wiener_index_sg(&sg, terminal, &info_sg);

      if (info.W != info_sg.W) {
        printf("n=%d bad, info.W=%Lu  info_sg.W=%Lu   terminal=%d\n",
               n, info.W, info_sg.W, terminal);
        if (++count_bad >= max_bad) assert(0);
      }
      if (n > 0 && info.diameter != g_diameter) {
        /* suspect diamstats() not made for n=0
           so don't worry how it matches code here */
        printf("n=%d bad, info.diameter=%d  g_diameter=%d   terminal=%d\n",
               n, info.diameter, g_diameter, terminal);
        if (++count_bad >= max_bad) assert(0);
      }
      if (info_sg.diameter != info.diameter) {
        printf("n=%d bad, info_sg.diameter=%d  info.diameter=%d   terminal=%d\n",
               n, info_sg.diameter, info.diameter, terminal);
        if (++count_bad >= max_bad) assert(0);
      }
      if (info.num != info_sg.num) {
        printf("n=%d bad, info.num=%d  info_sg.num=%d   terminal=%d\n",
               n, info.num, info_sg.num, terminal);
        if (++count_bad >= max_bad) assert(0);
      }
    }
  }

  {
    /* nautyextra_Wiener_index_sg() == nautyextra_Wiener_index_vpar() */
    int terminal;
    for (terminal = 0; terminal <= 1; terminal++) {
      struct nautyextra_Wiener_info info_sg;
      nautyextra_Wiener_index_sg(&sg, terminal, &info_sg);

      struct nautyextra_Wiener_info info_vpar;
      nautyextra_Wiener_index_vpar(vpar,n, terminal, &info_vpar);

      if (info_vpar.W != info_sg.W) {
        printf("n=%d bad, info_vpar.W=%Lu  info_sg.W=%Lu   terminal=%d\n",
               n, info_vpar.W, info_sg.W, terminal);
        printf("  %s", sgtog6(&sg));
        nautyextra_print_sg(&sg);
        nautyextra_view_sg(&sg);
        if (++count_bad >= max_bad) assert(0);
      }
      /* if (info_vpar.diameter != g_diameter) { */
      /*   printf("n=%d bad, info_vpar.diameter=%d  g_diameter=%d   terminal=%d\n", */
      /*          n, info_vpar.diameter, g_diameter, terminal); */
      /*   if (++count_bad >= max_bad) assert(0); */
      /* } */
      if (info_vpar.num != info_sg.num) {
        printf("n=%d Wiener index terminal=%d\n", n, terminal);
        printf("  info_vpar.num=%d\n", info_vpar.num);
        printf("  info_sg.num  =%d\n", info_sg.num);
        if (++count_bad >= max_bad) assert(0);
      }
    }
  }

  {
    /* nautyextra_diameter_vpar() == diamstats() */

    struct nautyextra_diameter_info  info;
    nautyextra_diameter_info_vpar(vpar,n, &info);
    if (n==0) {
      assert (info.end_v1 == 0);
      assert (info.end_v2 == 0);
      assert (info.end_ancestor == 0);
      assert (info.diameter == -1);
    } else {
      assert (1 <= info.end_v1 && info.end_v1 <= n);
      assert (1 <= info.end_v2 && info.end_v2 <= n);
      assert (1 <= info.end_ancestor && info.end_ancestor  <= n);
      int v1_depth = nautyextra_depth_of_vertex_vpar (vpar, info.end_v1);
      int v2_depth = nautyextra_depth_of_vertex_vpar (vpar, info.end_v2);
      int p_depth  = nautyextra_depth_of_vertex_vpar (vpar, info.end_ancestor);
      assert (info.diameter == v1_depth-p_depth + v2_depth-p_depth);
    }

    if (n > 0 && info.diameter != g_diameter) {
      /* suspect diamstats() not made for n=0
         so don't worry how it matches code here */
      printf("n=%d bad, diameter_vpar %d  g_diameter %d\n",
             n, info.diameter, g_diameter);
      nautyextra_print_sg(&sg);
      nautyextra_view_sg(&sg);
      if (++count_bad >= max_bad) assert(0);
    }
    if (info.radius != g_radius) {
      printf("n=%d bad, radius_vpar %d  g_radius %d\n",
             n, info.radius, g_radius);
      nautyextra_print_sg(&sg);
      nautyextra_view_sg(&sg);
      if (++count_bad >= max_bad) assert(0);
    }

    if (n <= 10) {
      int diameter_by_vpar = nautyextra_diameter_vpar(vpar,n);
      assert (info.diameter == diameter_by_vpar);

      int diameter_endpoints_by_ecc
        = test_diameter_endpoints_by_eccentricity(vpar,n);
      if (info.num_endpoints != diameter_endpoints_by_ecc) {
        printf ("diameter_endpoints_by_ecc  %d\n",
                diameter_endpoints_by_ecc);
        printf ("info.num_endpoints         %d\n",
                info.num_endpoints);
        nautyextra_print_vpar(vpar,n);
        nautyextra_view_synchronous = 1;
        nautyextra_view_vpar(vpar,n);
        if (++count_bad >= max_bad) assert(0);
      }

      int diameter_vertices_by_paths
        = test_diameter_vertices_by_paths(vpar,n);
      if (info.num_vertices != diameter_vertices_by_paths) {
        printf ("test_diameter_vertices_by_paths  %d\n",
                diameter_vertices_by_paths);
        printf ("info.num_vertices                %d\n",
                info.num_vertices);
        nautyextra_print_vpar(vpar,n);
        if (++count_bad >= max_bad) assert(0);
      }
    }
  }
}
void
my_gentreeg_outproc (FILE *outfile, int *orig_vpar, int n)
{
  int vpar[tree_max_vertices+1];
  assert (n <= tree_max_vertices);

  if (n==2 && count_trees[2]) return;  /* don't duplicate n=2 single tree */
  if (option_verbose) {
    printf("tree n=%d  ", n); nautyextra_print_vpar(vpar,n);
  }

  nautyextra_ranrelabel_vpar(orig_vpar,n,vpar);
  nautyextra_ranreroot_vpar(vpar,n);

  if (! (nautyextra_is_isomorphic_vpar2 (vpar,n, orig_vpar,n, 0) == TRUE)) {
    printf("n=%d randomized not isomorphic\n", n);
    printf("  orig_vpar "); nautyextra_print_vpar(orig_vpar,n);
    printf("  vpar      "); nautyextra_print_vpar(vpar,n);
    if (++count_bad >= max_bad) assert(0);
  }

  test_tree(vpar, n);
}
void
try_gentreeg (void)
{
  {
    int vpar[2];
    printf("gentreeg empty\n");
    vpar[0] = -1;
    vpar[1] = 0;
    my_gentreeg_outproc (stdout, vpar, 0);

    printf("gentreeg single\n");
    my_gentreeg_outproc (stdout, vpar, 1);
  }

  {
    printf("gentreeg\n");
    int num_vertices;
    for (num_vertices = tree_min_vertices;
         num_vertices <= tree_max_vertices;
         num_vertices++) {
      char num_vertices_str[20];
      sprintf(num_vertices_str, "%d", num_vertices);

      int gentreeg_argc = 0;
      char *gentreeg_argv[10];
      gentreeg_argv[gentreeg_argc++] = "gentreeg"; /* progname */
      /* gentreeg_argv[gentreeg_argc++] = "-q";       /\* quiet *\/ */
      gentreeg_argv[gentreeg_argc++] = num_vertices_str;
      gentreeg_argv[gentreeg_argc] = NULL;
      my_gentreeg_main(gentreeg_argc,gentreeg_argv);

      /* printf("%d %d\n", */
      /*        count_trees[num_vertices], */
      /*        count_connected_trees[num_vertices]); */
    }
  }
}


int count_cyclics[64];
int count_acyclics[64];
int total_triangles[64];
int count_domnum_by_d[64][64];
#define option_test_n9  1   /* a bit slow */
int indomnum_n9_max;
int indomnum_n9_graphs;
int indomnum_n9_count;
int totdomnum_n9_max;
int totdomnum_n9_graphs;

void
test_graph (const graph *g, int m, int n)
{
  assert (nautyextra_is_undirected(g,m,n) == TRUE);
  nautyextra_validate(g,m,n);
  count_total++;
  int is_cyclic = nautyextra_is_cyclic(g,m,n);
  int is_tree = ! is_cyclic;
  int is_connected = isconnected((graph*)g,m,n);

  if (option_verbose) {
    printf("graph %p m,n=%d,%d   cyclic=%d\n", g,m,n, is_cyclic);
    nautyextra_print_matrix(g,m,n);
    /* nautyextra_view(g,m,n); */
  }

  if (is_cyclic)
    count_cyclics[n]++;
  else
    count_acyclics[n]++;

  if (is_connected) {
    /* nautyextra_is_linegraph()
     */
    int by_func    = nautyextra_is_linegraph (g,m,n);
    int by_Beineke = test_is_linegraph_by_Beineke (g,m,n);
    if (by_func != by_Beineke) {
      printf("nautyextra_is_linegraph() n=%d\n", n);
      printf("  by_func     %d\n", by_func);
      printf("  by_Beineke  %d\n", by_Beineke);
      printf("%s", ntog6((graph*)g,m,n));
      nautyextra_view(g,m,n);
      if (++count_bad >= max_bad) assert(0);
    }
  }

  if (is_connected && n <= 7) {
    /* nautyextra_linegraph()
     */
    setword lg[graph_max_vertices*SETWORDSNEEDED(graph_max_vertices)];
    int ln = nautyextra_num_edges(g,m,n);
    int lm = SETWORDSNEEDED(ln);
    nautyextra_linegraph (g,m,n, lg,lm,ln);
    assert (nautyextra_is_undirected(lg,lm,ln) == TRUE);
    if (loopcount(lg,lm,ln)) {
      printf("nautyextra_linegraph() gives loops %d\n",
             loopcount(lg,lm,ln));
      nautyextra_print_matrix(lg,lm,ln);
      nautyextra_view(lg,lm,ln);
      if (++count_bad >= max_bad) assert(0);
    }
    nautyextra_validate(lg,lm,ln);
    assert (nautyextra_is_linegraph(lg,lm,ln) == TRUE);
  }

  if (is_connected && n >= 2) {
    /* nautyextra_is_cut_edge() vs isconnected()
     */
    int rep;
    for (rep = 1; rep <= 5; rep++) {
      int u, v;
      nautyextra_random_arc(g,m,n, &u,&v);
      assert (u>=0 && v>=0);
      int by_is_cut_edge = nautyextra_is_cut_edge(g,m,n, u,v);

      if (is_tree) {
        /* all edges in a tree are cut edges */
        if (! by_is_cut_edge) {
          printf("nautyextra_is_cut_edge() %d in a tree\n", by_is_cut_edge);
          nautyextra_view_synchronous = 1;
          nautyextra_view(g,m,n);
          if (++count_bad >= max_bad) assert(0);
        }
      }

      /* nautyextra_is_cut_edge() vs delete and isconnected() */
      assert(n <= graph_max_vertices);
      setword g2[graph_max_vertices*SETWORDSNEEDED(graph_max_vertices)];
      memcpy(g2, g, n*m*sizeof(g[0]));
      DELONEEDGE(g2, u,v, m);
      int by_isconnected = ! isconnected(g2,m,n);

      if (by_is_cut_edge != by_isconnected) {
        printf("nautyextra_is_cut_edge() doesn't match isconnected()\n");
        printf("  by_is_cut_edge    %d\n", by_is_cut_edge);
        printf("  by_isconnected    %d\n", by_isconnected);
        nautyextra_print_matrix(g,m,n);
        nautyextra_print_matrix(g2,m,n);
        nautyextra_view(g,m,n);
        nautyextra_view_synchronous = 1;
        nautyextra_view(g2,m,n);
        if (++count_bad >= max_bad) assert(0);
      }

      if (by_is_cut_edge) {
        assert (! isbiconnected((graph*)g,m,n));
      }
    }
  }

  if (is_connected) {  /* FIXME: nautyextra_is_cyclic() is only for connected */
    /* nautyextra_is_cyclic() vs girth() */
    int gg = girth((graph*)g,m,n);
    int is_cyclic_by_girth = (gg != 0);

    if (is_cyclic != is_cyclic_by_girth) {
      printf("nautyextra_is_cyclic() %d   n=%d isconnected=%d\n",
             is_cyclic, n, is_connected);
      printf("but by girth()=%d       %d\n", gg, is_cyclic_by_girth);
      if (++count_bad >= max_bad) assert(0);
    }
  }

  {
    /* nautyextra_triangle_count() vs numtriangles() */
    long                by_numtriangles = numtriangles((graph*)g,m,n);
    NAUTYEXTRA_TYPE_ULL by_nautyextra   = test_triangle_count(g,m,n);
    total_triangles[n] += by_numtriangles;

    if (by_nautyextra != by_numtriangles) {
      printf("test_triangle_count() doesn't match numtriangles()\n");
      if (++count_bad >= max_bad) assert(0);
    }

    if (n <= 8) {  /* bit slow */
      /* test_triangle_list_by_search() vs test_triangle_list_by_abc() */
      int list_by_search[3*1000];
      int list_by_abc[3*1000];
      int count_by_search = test_triangle_list_by_search
        (g,m,n, list_by_search, numberof(list_by_search)/3);
      if (count_by_search != by_numtriangles) {
        printf("test_triangle_list_by_search() count doesn't match numtriangles()\n");
        if (++count_bad >= max_bad) assert(0);
      }
      int count_by_abc = test_triangle_list_by_abc
        (g,m,n, list_by_abc, numberof(list_by_abc)/3);
      if (count_by_abc != by_numtriangles) {
        printf("test_triangle_list_by_abc() n=%d count doesn't match numtriangles()\n", n);
        printf("  test_triangle_list_by_abc() = %d\n", count_by_abc);
        printf("  numtriangles()              = %ld\n", by_numtriangles);
        if (++count_bad >= max_bad) assert(0);
      }
      if (memcmp(list_by_search, list_by_abc, 3*sizeof(int)*count_by_search)!= 0){
        printf("test_triangle_list_by_search() doesn't match test_triangle_list_by_abc()\n");
        int i;
        printf("by_search  ");
        for (i = 0; i < 3*count_by_search; i++)
          printf(" %d", list_by_search[i]);
        printf("\n");
        printf("by_abc     ");
        for (i = 0; i < 3*count_by_abc; i++)
          printf(" %d", list_by_abc[i]);
        printf("\n");
        if (++count_bad >= max_bad) assert(0);
      }
    }
  }

  {
    /* nautyextra_indnum_and_count() */
    struct nautyextra_indnum_and_count  info;
    nautyextra_indnum_and_count(g,m,n,&info);
    assert (info.indnum <= n);
    assert (info.indnum >= 0);
    assert (info.count < (1<<n));
  }

  SG_FREE(sg);     /* shrink so nde==elen */
  SG_INIT(sg);
  nauty_to_sg((graph*)g,&sg,m,n);

  int rep;
  for (rep = 1; rep <= n+1; rep++) {
    setword ss = test_random_setword(n);
    {
      /* nautyextra_is_domset1()
         nautyextra_is_domset_sg()
      */
      int by_test = test_is_domset_sg1 (&sg, ss);
      int by_g    = nautyextra_is_domset1 (g,n, ss);
      int by_sg   = nautyextra_is_domset_sg (&sg, &ss);
      if (by_sg != by_test || by_g != by_test) {
        printf("n=%d\n", n);
        nautyextra_print_matrix(g,m,n);
        nautyextra_print_sg(&sg);
        printf("ss %s\n", nautyextra_set_str(&ss,n));
        printf("test_is_domset_sg1()      %d\n", by_test);
        printf("nautyextra_is_domset1()   %d\n", by_g);
        printf("nautyextra_is_domset_sg() %d\n", by_sg);
        /* nautyextra_view_synchronous = 1; */
        /* nautyextra_view(g,m,n); */
      }
      assert (by_sg == by_test);
      assert (by_g == by_test);
    }
    {
      /* nautyextra_is_minimal_domset1()
         nautyextra_is_minimal_domset_sg()
      */
      int by_test = test_is_minimal_domset_sg1 (&sg, ss);
      int by_g    = nautyextra_is_minimal_domset1 (g,n, ss);
      int by_sg   = nautyextra_is_minimal_domset_sg (&sg, &ss);
      if (by_sg != by_test || by_g != by_test) {
        printf("n=%d\n", n);
        nautyextra_print_matrix(g,m,n);
        nautyextra_print_sg(&sg);
        printf("%s", sgtos6(&sg));
        printf("ss %s\n", nautyextra_set_str(&ss,n));
        printf("test_is_minimal_domset_sg1()      %d\n", by_test);
        printf("nautyextra_is_minimal_domset1()   %d\n", by_g);
        printf("nautyextra_is_minimal_domset_sg() %d\n", by_sg);
        /* nautyextra_view_synchronous = 1; */
        /* nautyextra_view(g,m,n); */
      }
      assert (by_sg == by_test);
      assert (by_g == by_test);
    }
  }
  {
    /* nautyextra_minimal_domsets_count1()
       nautyextra_minimal_domsets_count_sg()
    */
    unsigned long by_sg = nautyextra_minimal_domsets_count_sg (&sg);
    unsigned long by_g  = nautyextra_minimal_domsets_count1 (g,n);
    if (by_sg != by_g) {
      printf("n=%d\n", n);
      nautyextra_print_matrix(g,m,n);
      printf("nautyextra_minimal_domsets_count1()   %lu\n", by_g);
      printf("nautyextra_minimal_domsets_count_sg() %lu\n", by_sg);
    }
    assert (by_sg == by_g);
  }
  {
    /* nautyextra_domnum_and_count() */
    struct nautyextra_domnum_and_count  info;
    nautyextra_domnum_and_count_sg(&sg,&info);
    assert (info.domnum <= n);
    if (n==0 && ! (info.domnum==0)) {
      printf("nautyextra_domnum_and_count_sg() n=%d\n", n);
      printf("  domnum   %d want 0\n", info.domnum);
      printf("  count    %Lu\n", info.count);
      if (++count_bad >= max_bad) assert(0);
    }
    if (n>0 && ! (info.domnum > 0)) {
      printf("nautyextra_domnum_and_count_sg() n=%d\n", n);
      printf("  domnum   %d  want > 0\n", info.domnum);
      printf("  count    %Lu\n", info.count);
      if (++count_bad >= max_bad) assert(0);
    }
    count_domnum_by_d[n][info.domnum]++;
  }

  if (option_test_n9 && n==9 && is_connected) {
    {
      struct nautyextra_indomnum_and_count info;
      nautyextra_indomnum_and_count_sg(&sg,&info);
      if (info.indomnum > indomnum_n9_max) {
        indomnum_n9_max = info.indomnum;
        indomnum_n9_graphs = 0;
        indomnum_n9_count = 0;
      }
      if (info.indomnum == indomnum_n9_max) {
        indomnum_n9_graphs++;
        indomnum_n9_count += info.count;
      }
    }
    {
      struct nautyextra_totdomnum_and_count info;
      nautyextra_totdomnum_and_count_sg(&sg,&info);
      if (info.totdomnum > totdomnum_n9_max) {
        totdomnum_n9_max = info.totdomnum;
        totdomnum_n9_graphs = 0;
      }
      if (info.totdomnum == totdomnum_n9_max) {
        totdomnum_n9_graphs++;
      }
    }
  }
}
void
my_geng_outproc (FILE *outfile, graph *g, int n)
{
  setword g2[graph_max_vertices*SETWORDSNEEDED(graph_max_vertices)];
  int m = SETWORDSNEEDED(n);
  nautyextra_validate(g,m,n);
  nautyextra_ranrelabel(g,m,n,g2);
  test_graph(g2,m,n);
}
void
try_geng (void)
{
  DEBUG(printf("try_geng()\n"));

  static const int want_total_triangles[] = {
    /*   2 3 4  5                         */
    0,0, 0,1,7,48,401,4589,86317,2878659
  };
  memset(count_acyclics,'\0',sizeof(count_acyclics));
  memset(total_triangles,'\0',sizeof(total_triangles));

  printf("geng\n");
  /* option_verbose = 1; */

  int num_vertices;
  for (num_vertices = graph_min_vertices;
       num_vertices <= graph_max_vertices;
       num_vertices++) {
    char num_vertices_str[20];
    sprintf(num_vertices_str, "%d", num_vertices);

    int geng_argc = 0;
    char *geng_argv[10];
    geng_argv[geng_argc++] = "geng";
    /* geng_argv[geng_argc++] = "-q";     /\* quiet *\/ */
    geng_argv[geng_argc++] = "-c";     /* connected graphs */
    geng_argv[geng_argc++] = num_vertices_str;
    geng_argv[geng_argc] = NULL;
    my_geng_main(geng_argc,geng_argv);

    {
      static const int num_trees[] = {1,1,1,1,2,3,6,11,23,47,106,235,551,1301};
      if (num_vertices <= numberof(num_trees)) {
        if (count_acyclics[num_vertices] != num_trees[num_vertices]) {
          printf("count acyclics %d but num_trees %d\n",
                 count_acyclics[num_vertices],
                 num_trees[num_vertices]);
          if (++count_bad >= max_bad) assert(0);
        }
      }
    }

    if (num_vertices < numberof(want_total_triangles)) {
      if (total_triangles[num_vertices] != want_total_triangles[num_vertices]) {
        abort();
      }
    }

    if (option_test_n9 && num_vertices == 9) {
      if (indomnum_n9_max != 5) {
        printf("%d %d %d\n",
               indomnum_n9_max, indomnum_n9_graphs, indomnum_n9_count);
      }
      assert (indomnum_n9_max    == 5);
      assert (indomnum_n9_graphs == 1);
      assert (indomnum_n9_count  == 3);

      if (totdomnum_n9_max != 6) {
        printf("%d %d\n", totdomnum_n9_max, totdomnum_n9_graphs);
      }
      assert (totdomnum_n9_max    == 6);
      assert (totdomnum_n9_graphs == 2);
    }
  }
}

void
try_geng_unconnected (void)
{
  static const int want_domnum_counts[] = {
    /* A263284 */
    1,
    1, 1,
    2, 1, 1,
    4, 5, 1, 1,
    11, 16, 5, 1, 1,
    34, 94, 21, 5, 1, 1,
    156, 708, 152, 21, 5, 1, 1,
    1044, 9384, 1724, 166, 21, 5, 1, 1,
  };
  int want_domnum_counts_upto = 0;
  memset(count_domnum_by_d,'\0',sizeof(count_domnum_by_d));

  printf("geng unconnected\n");
  /* option_verbose = 1; */

  int num_vertices;
  for (num_vertices = ungraph_min_vertices;
       num_vertices <= ungraph_max_vertices;
       num_vertices++) {
    char num_vertices_str[20];
    sprintf(num_vertices_str, "%d", num_vertices);

    int geng_argc = 0;
    char *geng_argv[10];
    geng_argv[geng_argc++] = "geng";
    /* geng_argv[geng_argc++] = "-q";     /\* quiet *\/ */
    geng_argv[geng_argc++] = num_vertices_str;
    geng_argv[geng_argc] = NULL;
    my_geng_main(geng_argc,geng_argv);

    int d;
    for (d = 1; d <= num_vertices; d++) {
      if (want_domnum_counts_upto >= numberof(want_domnum_counts)) break;
      int want = want_domnum_counts[want_domnum_counts_upto++];
      int got  = count_domnum_by_d[num_vertices][d];
      if (want != got) {
        printf("domnum count by d   n=%d d=%d\n", num_vertices, d);
        printf("  got  %d\n", got);
        printf("  want  %d\n", want);
        if (++count_bad >= max_bad) assert(0);
      }
    }
  }
}

void
make_noduplicated_leaf_extremal_indnum (int *vpar, int n)
{
  int upto = 1;
  vpar[upto++] = 0;  /* root 1 */
  while (upto <= n) {
    int more = MIN(3, n-upto+1);
    int p = 1;
    int i;
    for (i=1; i<=more; i++) {
      vpar[upto] = p;
      p = upto;
      upto++;
    }
  }
  assert (upto==n+1);

}

/* Return the largest number of independent dominating sets in a connected
   graph of n vertices.
*/
unsigned long long
test_most_indomsets (int n)
{
  double d;
  int s = n/3;
  if ((n%3) == 0) {
    d = 2*pow(3.0,s-1) + pow(2.0,s-1);
  } else if ((n%3) == 1) {
    d = pow(3.0,s) + pow(2.0,s-1);
  } else {
    d = 4*pow(3.0,s-1) + 3.0*pow(2.0,s-2);
  }
  int c = (int) d;
  if (c != d) {
    DEBUG(printf("test_most_indomsets() n=%d  %g %d\n", n, d,c));
    abort();
  }
  return c;
}

void
test_seq_validate_vpar (const int *seq, int n)
{
  assert (n >= 0);
  int m = SETWORDSNEEDED(n+1);
  DYNALLSTAT(set,seen,seen_sz);
  DYNALLOC1(set,seen,seen_sz,m,"test_seq_validate");
  EMPTYSET(seen,m);

  if (seq[0] != 0) {
    printf("test_seq_validate_vpar() seq[0] = %d\n", seq[0]);
    abort();
  }
  int i;
  for (i=0; i<=n; i++) {
    int s = seq[i];
    assert (0 <= s && s <= n);
    assert (! ISELEMENT(seen,s));
    ADDELEMENT(seen,s);
  }
  assert(setsize(seen,m) == n+1);
}

void
try_various (void)
{
  DEBUG(printf("try_various()\n"));
  DYNALLSTAT(graph,g,g_sz);
  DYNALLSTAT(graph,g2,g2_sz);

  {
    /* nutyextra_Wiener_index() of n=0 vertices */
    struct nautyextra_Wiener_info info;
    nautyextra_Wiener_index (g,0,0,
                             0,  /* terminal */
                             &info);
    printf("nutyextra_Wiener_index() of n=0 vertices\n");
    printf("  num=%d diameter=%d from=%d to=%d\n",
           info.num, info.diameter,
           info.diameter_from, info.diameter_to);

    assert(info.W == 0);
    assert(info.num == 0);
    assert(info.diameter == -1);  /* no diameter ... */
    assert(info.diameter_from == -1);  /* no such vertex */
    assert(info.diameter_to   == -1);  /* no such vertex */
  }

  {
    /* Petersen
       https://hog.grinvin.org/ViewGraphInfo.action?id=660
    */
    static const char str[] = "IsP@OkWHG";
    int n = graphsize((char*)str);
    assert (n==10);
    int m = SETWORDSNEEDED(n);
    assert (m==1);
    DYNALLOC2(graph,g, g_sz, m,n,"try_various");
    stringtograph((char*)str,g,m);

    DYNALLOC2(graph,g2, g2_sz, m,n,"try_various");
    nautyextra_make_Petersen (g2,m,n, 2);
    assert(nautyextra_is_isomorphic(g,g2,m,n));

    /* not stable */
    assert (! nautyextra_is_stable(g,m,n));
    assert (nautyextra_max_stabilizing_sequence(g,m,n,NULL) == 1);
    complement(g,m,n);
    assert (nautyextra_max_stabilizing_sequence(g,m,n,NULL) == 1);
  }
  {
    /* Tree Semi-Stable
       Holton, "Completely Semi-Stable Trees"
       https://hog.grinvin.org/ViewGraphInfo.action?id=714

       #     1         4---0
       #      \       /            semi-stable at 3 and 4 only
       #       6 --- 5
       #      /       \
       #     2         3
       # >>graph6<<F?_ZG
    */
    static const char str[] = "F?_ZG";
    int n = graphsize((char*)str);
    assert (n==7);
    int m = SETWORDSNEEDED(n);
    assert (m==1);
    DYNALLOC2(graph,g, g_sz, m,n,"malloc");
    stringtograph((char*)str,g,m);

    static const int want[] = {0,0,0,1,1,0,0};
    assert (numberof(want) == n);
    int v;
    for (v = 0; v < n; v++) {
      assert (nautyextra_is_semistable_at_vertex(g,m,n,v) == want[v]);
    }
  }
  {
    /* Octahedral Graph, n=6
     */
    static const char str[] = "E}lw";
    int n = graphsize((char*)str);
    assert (n==6);
    int m = SETWORDSNEEDED(n);
    assert (m==1);
    DYNALLOC2(graph,g, g_sz, m,n,"malloc");
    stringtograph((char*)str,g,m);

    struct nautyextra_domnum_and_count info;
    nautyextra_domnum_and_count1(g, n, &info);
    assert (info.domnum == 2);
    assert (info.count == 15);
    assert (nautyextra_minimal_domsets_count1(g,n) == 15);
  }
  {
    /* Most Minimal Domsets, N=9
     */
    static const struct {
      const char    *str;
      int           want_num_edges;
      int           want_domnum;
      unsigned long want_domnum_count;
      unsigned long want_minimal_domsets_count;
    } elems[] = {
      { ":H`A_wa_K`WaEQG`cJ",  18, 3, 51, 51 },
      { ":H`AK?qFG[@cIYGaDL",  19, 2,  1, 51 },
      { ":H`AK@qEQOqPGSsPDK^", 20, 2,  1, 51 }
    };
    int i;
    for (i=0; i<numberof(elems); i++) {
      const char *str = elems[i].str;
      int n = graphsize((char*)str);
      assert (n==9);
      int m = SETWORDSNEEDED(n);
      assert (m==1);
      DYNALLOC2(graph,g, g_sz, m,n,"malloc");
      stringtograph((char*)str,g,m);

      assert (nautyextra_num_edges(g,m,n) == elems[i].want_num_edges);
      struct nautyextra_domnum_and_count info;
      nautyextra_domnum_and_count1(g, n, &info);
      printf ("%d %Lu %Lu\n", info.domnum, info.count,
              nautyextra_minimal_domsets_count1(g,n));
      assert (info.domnum == elems[i].want_domnum);
      assert (info.count  == elems[i].want_domnum_count);
      assert (nautyextra_minimal_domsets_count1(g,n)
              == elems[i].want_minimal_domsets_count);
    }
  }

  {
    /* nautyextra_is_matching_vpar_set() on n=0,
       empty set is a matching
    */
    static const set ss[] = {0};
    assert (nautyextra_is_matching_vpar_set(NULL,0, ss) == 1);
    assert (nautyextra_is_matching_vpar_set(NULL,0, NULL) == 1);
    assert (nautyextra_is_matching_vpar_set(vpar_empty,0, ss) == 1);
    assert (nautyextra_is_matching_vpar_set(vpar_empty,0, NULL) == 1);
  }
  {
    /* NAUTYEXTRA_SET1_LOOP() on n=0 */
    int count = 0;
    setword last_ss = 123;
    setword ss;
    NAUTYEXTRA_SET1_LOOP(ss,0, count++; last_ss = ss);
    assert (count == 1);
    assert (last_ss == 0);
  }
  {
    /* nautyextra_matchnum_and_count_vpar() */
    struct nautyextra_matchnum_and_count  info;
    nautyextra_matchnum_and_count_vpar (vpar_singleton,1, &info);
    assert (info.matchnum == 0);
    assert (info.count == 1);
  }
  {
    /* Different list orders in sg are preserved in sgtos6().
       Have nautyextra_canonise_sg() sort them for consistency.
    */
    const char str1[] = ":a_OWSEH@bOXCATJdRXS}cPASi\\IjQjsjLEv[lvj]B";
    const char str2[] = ":a_OWSEH@bOXCATJdRXS}cPBSi\\IjQjsjLEv[lvj]B";
    int nloops;
    stringtosparsegraph((char*)str1,&sg,&nloops);
    assert(nloops==0);
    stringtosparsegraph((char*)str2,&sg2,&nloops);
    assert(nloops==0);
    assert(nautyextra_is_isomorphic_sg(&sg,&sg2));

    nautyextra_canonise_sg(&sg,&canon_sg);
    nautyextra_canonise_sg(&sg2,&canon_sg2);
    assert(! aresame_sg(&sg,&sg2));
    assert(  aresame_sg(&canon_sg,&canon_sg2));
    {
      char *canon_str1 = strdup(sgtog6(&canon_sg));
      char *canon_str2 = strdup(sgtog6(&canon_sg2));
      /* printf("%s",canon_str1); */
      /* printf("%s",canon_str2); */
      assert(strcmp(canon_str1,canon_str2)==0);
      free(canon_str1);
      free(canon_str2);
    }
  }
  {
    /* vpar_is_isomorphic_vpar() samples */
    struct elem {
      int want;
      int n;
      int vpar1[11];
      int vpar2[11];
    } elems[] = {
      /* forests, different heights */
      /*              1 2 3 4 5  6 7 8  */
      { FALSE, 8, {0, 0,1,2,1,4, 0,5,4}, {0, 0,1,1,3, 0,4,5,4} },

      { TRUE, 0, {0}, {0} },                   /* empty */
      { TRUE, 1, {0, 0}, {0, 0} },             /* single vertex */
      { FALSE, 3, {0, 0,1,1}, {0, 0,1,2} },    /* path-3 vs star-3 */
      { TRUE,  4, {0, 0,1,2,1}, {0, 0,1,1,3} },  /* diff numbering */

      /* forests, same heights */
      /*              1 2 3 4  5 6 7 8       1 2 3 4  5 6 7 8  */
      { TRUE,  8, {0, 0,1,2,1, 0,5,6,5}, {0, 0,1,2,1, 0,5,6,5} },
      { FALSE, 8, {0, 0,1,2,1, 0,5,6,5}, {0, 0,1,2,2, 0,5,6,5} },
      { FALSE, 8, {0, 0,1,2,1, 0,5,6,5}, {0, 0,1,2,1, 0,5,6,6} },

    };
    int i;
    for (i=0; i<numberof(elems); i++) {
      int want = elems[i].want;
      int n    = elems[i].n;
      const int *vpar1 = elems[i].vpar1;
      const int *vpar2 = elems[i].vpar2;

      int  got_perm[tree_max_vertices+1];
      int got = nautyextra_is_isomorphic_vpar(vpar1,vpar2,n, 1, got_perm); /*directed*/
      if (got != want) {
        printf("nautyextra_is_isomorphic_vpar() n=%d got %d want %d\n",
               n, got, want);
        nautyextra_print_vpar(vpar1,n);
        nautyextra_print_vpar(vpar2,n);
        /* nautyextra_view_vpar(vpar1,n); */
        /* nautyextra_view_synchronous = 1; */
        /* nautyextra_view_vpar(vpar2,n); */
        if (++count_bad >= max_bad) assert(0);

      } else if (got) {
        int  vpar2_again[tree_max_vertices+1];
        nautyextra_relabel_perm_vpar (vpar1, n, got_perm, vpar2_again);
        /* printf("vpar2_again "); nautyextra_print_vpar(vpar2_again,n); */

        if (! nautyextra_aresame_vpar(vpar2, vpar2_again, n)) {
          printf("nautyextra_is_isomorphic_vpar() n=%d perm relabel not same\n",
                 n);
          if (++count_bad >= max_bad) assert(0);
        }
      }
    }
  }
  {
    /* nautyextra_subtree_heights_vpar() in sequence */
    const int max_n = 10;
    int vpar[max_n+1];
    int seq[max_n+1];

    int n;
    for (n=0; n<=max_n; n++) {
      int rep;
      for (rep=1; rep < n+1; rep++) {
        nautyextra_make_random_vpar(vpar,n, 1);  /* forest */

        DYNALLSTAT(int,heights,heights_sz);
        DYNALLOC1(int,heights,heights_sz,n+1,"try_various");
        nautyextra_subtree_heights_vpar (vpar,n, heights);

        memset(seq,'\xFF',sizeof(seq));
        nautyextra_seq_subtree_height_vpar (vpar,n, seq);
        DEBUG({ printf("seq ");
                int i;
                for(i=0;i<=n;i++) printf("%d,",seq[i]);
                printf("\n");
              });
        test_seq_validate_vpar (seq,n);

        int v;
        for (v=1; v<=n; v++) {
          int s = seq[v];
          assert (1 <= s && s <= n);
          if (v < n) {
            int t = seq[v+1];
            assert (1 <= t && t <= n);
            int sh = heights[s];
            int th = heights[t];
            assert (sh <= th);
          }
        }
      }
    }
  }
  {
    int n = 5;
    int m = 1;
    setword g[n*m];
    EMPTYGRAPH(g,m,n);
    assert (! ISELEMENT(GRAPHROW(g,1,m),4));
    assert (! ISELEMENT(GRAPHROW(g,4,m),1));
    ADDONEARC0(g, 1,4, m);
    assert (ISELEMENT(GRAPHROW(g,1,m),4));
    assert (! ISELEMENT(GRAPHROW(g,4,m),1));
    DELONEARC0(g, 4,1, m);
    assert (ISELEMENT(GRAPHROW(g,1,m),4));
    assert (! ISELEMENT(GRAPHROW(g,4,m),1));
    DELONEARC0(g, 1,4, m);
    assert (! ISELEMENT(GRAPHROW(g,1,m),4));
    assert (! ISELEMENT(GRAPHROW(g,4,m),1));

    ADDONEARC1(g, 1,4, m_not_used);
    assert (ISELEMENT(GRAPHROW(g,1,m),4));
    assert (! ISELEMENT(GRAPHROW(g,4,m),1));
    DELONEARC1(g, 4,1, m_not_used);
    assert (ISELEMENT(GRAPHROW(g,1,m),4));
    assert (! ISELEMENT(GRAPHROW(g,4,m),1));
    DELONEARC1(g, 1,4, m_not_used);
    assert (! ISELEMENT(GRAPHROW(g,1,m),4));
    assert (! ISELEMENT(GRAPHROW(g,4,m),1));
  }

  {
    DEBUG(printf("nautyextra_make_random_vpar() validate\n"));
    const int max_n = 10;
    int vpar[max_n+1];
    int n;
    for (n=0; n<=max_n; n++) {
      nautyextra_make_random_vpar(vpar,n, 0);
      nautyextra_validate_vpar(vpar,n);

      nautyextra_make_random_vpar(vpar,n, 1);
      nautyextra_validate_vpar(vpar,n);
    }
  }
  {
    /* inverse domination Kulli and Sigarkanti example */
    sg.nv = 0;
    nautyextra_add_edge_sg(&sg, 1-1, 2-1);
    nautyextra_add_edge_sg(&sg, 1-1, 5-1);
    nautyextra_add_edge_sg(&sg, 2-1, 3-1);
    nautyextra_add_edge_sg(&sg, 1-1, 5-1);
    nautyextra_add_edge_sg(&sg, 4-1, 5-1);
    int n = sg.nv;
    assert (SETWORDSNEEDED(n) <= 1);
    set ss[1];

    ss[0]=0;
    ADDELEMENT(ss, 2-1);
    ADDELEMENT(ss, 5-1);
    assert(nautyextra_is_invdomset_sg(&sg,ss));

    /* invdomnum three sets  2,5  2,4  3,5 */
    struct nautyextra_invdomnum_and_count  info;
    nautyextra_invdomnum_and_count_sg (&sg, &info);
    assert(info.invdomnum == 2);
    assert(info.count == 3);

    NAUTYEXTRA_TYPE_ULL got = nautyextra_invdomsets_count_sg(&sg);
    printf("%Lu\n",got);
    assert(got==6);
  }

  {
    /* vpar_is_isomorphic_vpar() vs vpar_is_isomorphic_vpar2(), random */
    const int max_n = 10;
    int vpar[max_n+1];
    int vpar2[max_n+1];
    int n;
    for (n=0; n<=max_n; n++) {
      int num_reps = (n <= 4 ? n+1 : 10*n*n);
      int rep;
      for (rep=1; rep < num_reps; rep++) {
        nautyextra_make_random_vpar(vpar,n, rep&1);  /* tree or forest */
        nautyextra_make_random_vpar(vpar2,n, rep&1);
        int want = nautyextra_is_isomorphic_vpar2(vpar,n, vpar2,n, 1);
        int got = nautyextra_is_isomorphic_vpar(vpar,vpar2,n, 1, NULL); /*directed*/
        if (got != want) {
          printf("nautyextra_is_isomorphic_vpar() random got %d want %d\n",
                 got, want);
          nautyextra_print_vpar(vpar,n);
          nautyextra_print_vpar(vpar2,n);
          /* nautyextra_view_vpar(vpar,n); */
          /* nautyextra_view_synchronous = 1; */
          /* nautyextra_view_vpar(vpar2,n); */
          if (++count_bad >= max_bad) assert(0);
        }
      }
    }
  }
  {
    /* vpar_is_isomorphic_vpar() vs ranrelabel */
    const int max_n = 10;
    int vpar[max_n+1];
    int vpar2[max_n+1];
    int n;
    for (n=0; n<=max_n; n++) {
      int rep;
      for (rep=1; rep < n+1; rep++) {
        nautyextra_make_random_vpar(vpar,n, 1);  /* forest */
        nautyextra_ranrelabel_vpar(vpar,n,vpar2);

        int got_perm[max_n+1];
        if (! (nautyextra_is_isomorphic_vpar(vpar,vpar2,n, 1, got_perm)
               == TRUE)) {
          nautyextra_view_vpar(vpar,n);
          nautyextra_view_vpar(vpar2,n);
          printf("nautyextra_is_isomorphic_vpar() of ranrelabel should be TRUE\n");
          if (++count_bad >= max_bad) assert(0);

        } else {
          int  vpar2_again[tree_max_vertices+1];
          nautyextra_relabel_perm_vpar (vpar, n, got_perm, vpar2_again);
          if (! nautyextra_aresame_vpar(vpar2_again, vpar2, n)) {
            printf("nautyextra_is_isomorphic_vpar() of ranrelabel perm does not restore\n");
            if (++count_bad >= max_bad) assert(0);
          }
        }
      }
    }
  }
  {
    /* nautyextra_seq_upascend() samples */
    const int n = 3;
    const int vpar[] = { 0, 3,0,0 };
    {
      const int want[] = { 1,2,3, 12345 };
      int got[4];
      got[3] = 12345;
      nautyextra_seq_upascend(vpar,n,0, got);
      assert (memcmp(got, want, 4*sizeof(int)) == 0);
    }
    {
      const int want[] = { 1,3,2, 12345 };
      int got[4];
      got[3] = 12345;
      nautyextra_seq_upascend(vpar,n,1, got);
      assert (memcmp(got, want, 4*sizeof(int)) == 0);
    }
  }
  {
    /* nautyextra_from_upascend() samples */
    const int n = 4;
    const int vec[] = { 4,3,0,0 };
    {
      const int want[] = { 0, 4,3,0,0, 12345 };
      assert(numberof(want)==n+2);
      int got[6];
      got[5] = 12345;
      nautyextra_from_upascend(vec,n,0, got);
      if (! nautyextra_aresame_vpar(got, want, n)) {
        DEBUG(printf("ahead=0 got  ");nautyextra_print_vpar(got,n));
        DEBUG(printf("       want  ");nautyextra_print_vpar(want,n));
        if (++count_bad >= max_bad) assert(0);
      }
    }
    {
      const int want[] = { 0, 4,0,0,3, 12345 };
      assert(numberof(want)==n+2);
      int got[6];
      got[5] = 12345;
      nautyextra_from_upascend(vec,n,1, got);
      if (! nautyextra_aresame_vpar(got, want, n)) {
        DEBUG(printf("ahead=0 got  ");nautyextra_print_vpar(got,n));
        if (++count_bad >= max_bad) assert(0);
      }
    }
  }
  {
    /* upascend */
    int max_n = 3;
    int forest;
    for (forest = 0; forest <= 1; forest++) {
      int n;
      for (n = 0; n <= max_n; n++) {
        int vec[max_n];
        memset(vec,'\xFF',sizeof(vec));
        int count = 0;
        nautyextra_can_init(vec,n, forest);
        do {
          count++;
          int ahead;
          for (ahead = 0; ahead <= 1; ahead++) {
            DEBUG({ printf("upascend vec ");
                    int i;
                    for(i=0;i<n;i++) printf("%d,",vec[i]);
                    printf("\n");
                  });

            int vpar[max_n+1];
            nautyextra_from_upascend(vec,n,ahead, vpar);
            DEBUG(printf("  ");nautyextra_print_vpar(vpar,n));
            nautyextra_validate_vpar(vpar,n);

            int seq_again[max_n];
            nautyextra_seq_upascend(vpar,n,ahead, seq_again);

            int vec_again[max_n];
            int i;
            for (i=0;i<n;i++) vec_again[i]=vpar[seq_again[i]];

            if (memcmp(vec, vec_again, n*sizeof(int)) != 0) {
              printf ("upascend vec_again different\n");
              printf("vec_again ");
              int i;
              for(i=0;i<n;i++) printf("%d,",vec[i]);
              printf("\n");
              if (++count_bad >= max_bad) assert(0);
            }
          }
        } while (nautyextra_can_next(vec,n, forest));

        {
          int want_count = (n==0 ? 1 : pow((n+forest), n-1));
          if (count != want_count) {
            printf ("n=%d count %d want %d\n", n, count, want_count);
            if (++count_bad >= max_bad) assert(0);
          }
        }
      }
    }
  }
  {
    /* path-n by vpar */
    /*                                                0  1  2  3  4  5  6  7 */
    static const int want_domnum[]                = { 0, 1, 1, 1, 2, 2, 2, 3 };
    static const int want_domnum_count[]          = { 1, 1, 2, 1, 4, 3, 1, 8 };
    static const int want_minimal_domsets_count[] = { 1, 1, 2, 2, 4, 4, 7, 9, 13, 18, 25, 36 };
    static const int want_totdomnum[]             = { 0,-1, 2, 2, 2, 3, 4, 4, 4 };
    static const int want_indomnum[]              = { 0, 1, 1, 1, 2, 2, 2, 3, 3 };
    static const int max_n = 20;
    int vpar[max_n + 1];
    int max_m = SETWORDSNEEDED(max_n);
    graph g[max_n*max_m];
    int m,n;
    for (n = 0; n <= max_n; n++) {
      nautyextra_make_path_vpar (vpar,n);
      nautyextra_vpar_to_sg(vpar,n, &sg, 0);
      m = SETWORDSNEEDED(n);
      nautyextra_vpar_to_graph(vpar, g,m,n, 0);
      /* printf("path g "); nautyextra_print_matrix (g,m,n); */
      if (n>=1) assert (isconnected(g,m,n) == TRUE);
      assert (nautyextra_is_connected_vpar(vpar,n) == TRUE);

      if (n <= 8) {
        int got  = nautyextra_bandwidth_vpar(vpar,n);
        int want = (n<=1 ? 0 : 1);
        if (got != want) {
          printf("path n=%d nautyextra_bandwidth_vpar() got %d want %d\n",
                 n,got,want);
          if (++count_bad >= max_bad) assert(0);
        }
      }
      {
        int got  = nautyextra_height_vpar(vpar,n);
        int want = n-1;
        if (got != want) {
          printf("path n=%d nautyextra_height_vpar() got %d want %d\n",
                 n,got,want);
          if (++count_bad >= max_bad) assert(0);
        }
      }
      {
        int got  = nautyextra_2height_vpar(vpar,n);
        int want = (n==0 ? -1 : n-2);
        if (got != want) {
          printf("path n=%d nautyextra_2height_vpar() got %d want %d\n",
                 n,got,want);
          if (++count_bad >= max_bad) assert(0);
        }
      }
      {
        int got  = nautyextra_limb_length_vpar(vpar,n);
        int want = (n<=1 ? 0 : n-1);
        if (got != want) {
          printf("path n=%d nautyextra_limb_length_vpar() got %d want %d\n",
                 n,got,want);
          nautyextra_view_vpar(vpar,n);
          if (++count_bad >= max_bad) assert(0);
        }
      }
      {
        int got  = nautyextra_limb_height_vpar(vpar,n);
        int want = (n==0 ? 0 : n-1);
        if (got != want) {
          printf("path n=%d nautyextra_limb_height_vpar() got %d want %d\n",
                 n,got,want);
          if (++count_bad >= max_bad) assert(0);
        }
      }

      {
        /* nautyextra_diameter_info_vpar() of path */
        struct nautyextra_diameter_info  info;
        nautyextra_diameter_info_vpar (vpar,n, &info);
        assert (info.diameter == n-1);
        assert (info.num_endpoints == (n < 2 ? n : 2));
      }
      {
        /* nautyextra_matchnum_and_count_vpar() of path */
        int by_floor       = n >> 1;
        int by_floor_count = (n&1 ? (n>>1)+1 : 1);
        struct nautyextra_matchnum_and_count  info;
        nautyextra_matchnum_and_count_vpar (vpar,n, &info);
        if (info.matchnum != by_floor || info.count != by_floor_count) {
          nautyextra_print_vpar(vpar,n);
          printf("path n=%d matchnum_and_count\n", n);
          printf("  by_vpar   matchnum %d count %Lu\n",
                 info.matchnum, info.count);
          printf("  by floor  matchnum %d count %d\n",
                 by_floor, by_floor_count);
          if (++count_bad >= max_bad) assert(0);
        }
      }
      {
        /* nautyextra_matchnum_vpar() */
        int by_floor = n >> 1;
        int by_vpar  = nautyextra_matchnum_vpar (vpar, n);
        if (by_floor != by_vpar) {
          printf("path n=%d matchnum by_vpar %d by_floor %d\n",
                 n,by_vpar,by_floor);
          nautyextra_print_vpar(vpar,n);
          if (++count_bad >= max_bad) assert(0);
        }
      }

      {
        /* nautyextra_minimal_domsets_count_vpar() */
        int got_count = nautyextra_minimal_domsets_count_vpar(vpar,n);
        if (n < numberof(want_minimal_domsets_count) && got_count != (want_minimal_domsets_count)[n]) {
          printf("nautyextra_minimal_domsets_count_vpar() n=%d path count=%d want %d\n",
                 n, got_count, want_minimal_domsets_count[n]);
          nautyextra_print_vpar(vpar,n);
          if (++count_bad >= max_bad) assert(0);
        }
      }
      {
        /* nautyextra_minimal_domsets_count1() */
        int got_count = nautyextra_minimal_domsets_count1(g,n);
        if (n < numberof(want_minimal_domsets_count) && got_count != (want_minimal_domsets_count)[n]) {
          printf("nautyextra_minimal_domsets_count1() n=%d path count=%d want %d\n",
                 n, got_count, want_minimal_domsets_count[n]);
          if (++count_bad >= max_bad) assert(0);
        }
      }
      {
        /* nautyextra_minimal_domsets_count_sg() */
        int got_count = nautyextra_minimal_domsets_count_sg(&sg);
        if (n < numberof(want_minimal_domsets_count) && got_count != (want_minimal_domsets_count)[n]) {
          printf("nautyextra_minimal_domsets_count_sg() n=%d path count=%d want %d\n",
                 n, got_count, want_minimal_domsets_count[n]);
          if (++count_bad >= max_bad) assert(0);
        }
      }

      {
        /* nautyextra_domnum_vpar() */
        int got_domnum = nautyextra_domnum_vpar(vpar,n);
        if (n < numberof(want_domnum) && got_domnum != want_domnum[n]) {
          printf("nautyextra_domnum_vpar() n=%d path domnum=%d want %d\n",
                 n, got_domnum, want_domnum[n]);
          nautyextra_print_vpar(vpar,n);
          if (++count_bad >= max_bad) assert(0);
        }
      }
      if (n <= WORDSIZE) {
        /* nautyextra_domnum_and_count1() */
        struct nautyextra_domnum_and_count info;
        nautyextra_domnum_and_count1(g, n, &info);
        if ((n < numberof(want_domnum) &&
             info.domnum != want_domnum[n])
            || (n < numberof(want_domnum_count) &&
                info.count != want_domnum_count[n])) {
          printf("nautyextra_domnum_and_count1() n=%d path\n", n);
          printf("  by_g   domnum %d count %Lu\n", info.domnum, info.count);
          printf("  data   domnum %d count %d\n", want_domnum[n], want_domnum_count[n]);
          nautyextra_print_vpar(vpar,n);
          if (++count_bad >= max_bad) assert(0);
        }
      }
      {
        /* nautyextra_domnum_and_count_sg() */
        struct nautyextra_domnum_and_count info;
        nautyextra_domnum_and_count_sg(&sg, &info);
        if ((n < numberof(want_domnum) &&
             info.domnum != want_domnum[n])
            || (n < numberof(want_domnum_count) &&
                info.count != want_domnum_count[n])) {
          printf("nautyextra_domnum_and_count_sg() n=%d path\n", n);
          printf("  by_sg  domnum %d count %Lu\n", info.domnum, info.count);
          printf("  data   domnum %d count %d\n", want_domnum[n], want_domnum_count[n]);
          nautyextra_print_vpar(vpar,n);
          if (++count_bad >= max_bad) assert(0);
        }
      }
      {
        /* nautyextra_domnum_and_count_vpar() */
        struct nautyextra_domnum_and_count info;
        nautyextra_domnum_and_count_vpar(vpar,n, &info);
        if (n < numberof(want_domnum) && info.domnum != want_domnum[n]) {
          printf("nautyextra_domnum_and_count_vpar() n=%d path\n", n);
          printf("  got  domnum %d count %Lu\n", info.domnum, info.count);
          printf("  want domnum %d count %d\n", want_domnum[n], want_domnum_count[n]);
          nautyextra_print_vpar(vpar,n);
          if (++count_bad >= max_bad) assert(0);
        }
      }
      {
        /* nautyextra_totdomnum_vpar() */
        int got_totdomnum = nautyextra_totdomnum_vpar(vpar,n);
        if (n < numberof(want_totdomnum)
            && got_totdomnum != want_totdomnum[n]) {
          nautyextra_print_vpar(vpar,n);
          printf("nautyextra_totdomnum_vpar() n=%d path\n", n);
          printf("  by_vpar %d\n", got_totdomnum);
          printf("  data    %d\n", want_domnum[n]);
          if (++count_bad >= max_bad) assert(0);
        }

        struct nautyextra_totdomnum_and_count info_sg;
        nautyextra_totdomnum_and_count_sg(&sg, &info_sg);

        struct nautyextra_totdomnum_and_count info_vpar;
        nautyextra_totdomnum_and_count_vpar(vpar,n, &info_vpar);

        if (info_vpar.totdomnum != info_sg.totdomnum
            || info_vpar.count != info_sg.count) {
          printf("path %d totdomnum_and_count\n", n);
          printf("  info_vpar  %d count %Lu\n",
                 info_vpar.totdomnum, info_vpar.count);
          printf("  info_sg    %d count %Lu\n",
                 info_sg.totdomnum, info_sg.count);
          nautyextra_view_vpar(vpar,n);
          if (++count_bad >= max_bad) assert(0);
        }
      }
      {
        struct nautyextra_indomnum_and_count info_sg;
        nautyextra_indomnum_and_count_sg(&sg, &info_sg);

        struct nautyextra_indomnum_and_count info_vpar;
        nautyextra_indomnum_and_count_vpar(vpar,n, &info_vpar);

        if (info_vpar.indomnum != info_sg.indomnum
            || info_vpar.count != info_sg.count) {
          printf("path %d indomnum_and_count\n", n);
          printf("  info_vpar  %d count %Lu\n",
                 info_vpar.indomnum, info_vpar.count);
          printf("  info_sg    %d count %Lu\n",
                 info_sg.indomnum, info_sg.count);
          nautyextra_print_vpar(vpar,n);
          nautyextra_view_vpar(vpar,n);
          if (++count_bad >= max_bad) assert(0);
        }

        /* nautyextra_indomnum_vpar() */
        int got_indomnum = nautyextra_indomnum_vpar(vpar,n);
        if (n < numberof(want_indomnum)
            && got_indomnum != want_indomnum[n]) {
          nautyextra_print_vpar(vpar,n);
          printf("nautyextra_indomnum_vpar() n=%d path\n", n);
          printf("  by_vpar %d\n", got_indomnum);
          printf("  data    %d\n", want_domnum[n]);
          if (++count_bad >= max_bad) assert(0);
        }
      }
    }
  }
  {
    /* nautyextra_is_distance_regular() */
    const int n_max = 7;
    int seen[1*2*3*4*5*6*7][n_max+1];
    int n;
    for (n = 2; n <= 7; n++) {
      unsigned long long f = nautyextra_factorial(n);
      /* printf("numtoperm1 %Lu\n", f); */
      unsigned long long num;
      for (num = 0; num < f; num++) {
        nautyextra_numtoperm1(num,n, &seen[num][0]);
        int i;
        for (i = 0; i < num; i++) {
          if (memcmp(&seen[i][1], &seen[num][0], n*sizeof(int)) == 0) {
            printf ("same\n");
            if (++count_bad >= max_bad) assert(0);
          }
        }
      }
    }
  }
  {
    /* nautyextra_is_distance_regular() */
    const int n = 4;
    const int m = SETWORDSNEEDED(n);
    assert (m == 1);
    graph g[n];

    nautyextra_make_complete (g,m,n, 0);
    assert(loopcount(g,m,n) == 0);
    nautyextra_validate(g,m,n);

    assert (nautyextra_is_distance_regular(g,m,n, NULL) == TRUE);
  }
  {
    /* diamond */
    const int n = 4;
    const int m = SETWORDSNEEDED(n);
    assert (m==1);
    setword g[n];
    EMPTYGRAPH(g,m,n);
    ADDONEEDGE(g, 0,1, m);
    ADDONEEDGE(g, 0,2, m);
    ADDONEEDGE(g, 0,3, m);
    ADDONEEDGE(g, 1,3, m);
    ADDONEEDGE(g, 2,3, m);
    setword g2[n];
    nautyextra_ranrelabel(g,m,n,g2);
    assert (nautyextra_diamond_count(g2,m,n, 0) == 1);
    assert (nautyextra_diamond_count(g2,m,n, 1) == 1);
  }
  {
    /* nautyextra_diamond_count() */
    int n = 4;
    int m = SETWORDSNEEDED(n);
    assert (m == 1);
    graph g[5];
    EMPTYGRAPH(g,m,5);

    nautyextra_make_complete (g,m,n, 0);
    assert(loopcount(g,m,n) == 0);
    nautyextra_validate(g,m,n);

    assert (nautyextra_diamond_count(g,m,n, 0) == 6);
    assert (nautyextra_binomial(4,2) == 6);
    assert (nautyextra_diamond_count(g,m,n, 1) == 0);

    /* with extra leaf */
    n++;
    m = SETWORDSNEEDED(n);
    assert (m == 1);
    ADDONEEDGE(g, 0,4, m);
    /* nautyextra_view(g,m,n); */
    nautyextra_validate(g,m,n);

    assert (nautyextra_diamond_count(g,m,n, 0) == 6);
    assert (nautyextra_diamond_count(g,m,n, 1) == 0);
  }
  {
    /* triangle line graph = triangle */
    const int n = 3;
    const int m = SETWORDSNEEDED(n);
    assert (m==1);
    setword g1[m],g2[m];
    nautyextra_make_cycle (g1,m,n, 0);
    assert (nautyextra_is_linegraph(g1,m,n) == TRUE);

    nautyextra_linegraph (g1,m,n, g2,m,n);
    nautyextra_validate(g2,m,n);
    /* nautyextra_view(g1,m,n); */
    /* nautyextra_view(g2,m,n); */
    assert (nautyextra_is_isomorphic(g1,g2,m,n) == TRUE);
  }
  {
    int m = SETWORDSNEEDED(test_make_Beineke_maxn);
    assert (m==1);
    setword g[test_make_Beineke_maxn];
    int num;
    for (num = 1; num <= 9; num++) {
      DEBUG(printf("Beineke G%d\n", num));
      int n = test_make_Beineke(g,m, num);
      nautyextra_validate(g,m,n);
      assert (loopcount(g,m,n) == 0);
      int by_func = nautyextra_is_linegraph (g,m,n);
      if (by_func != FALSE) {
        printf("nautyextra_is_linegraph() Beineke G%d\n", num);
        printf("  by_func     %d\n", by_func);
        /* nautyextra_view(g,m,n); */
        printf("%s", ntog6((graph*)g,m,n));
        if (++count_bad >= max_bad) assert(0);
      }
      int by_Beineke = test_is_linegraph_by_Beineke (g,m,n);
      if (by_Beineke != FALSE) {
        printf("nautyextra_is_linegraph() Beineke G%d\n", num);
        printf("  by_Beineke %d\n", by_func);
        nautyextra_view(g,m,n);
        if (++count_bad >= max_bad) assert(0);
      }
      /* if (by_func != by_Beineke) { */
      /*   printf("nautyextra_is_linegraph() n=%d\n", n); */
      /*   printf("  by_func     %d\n", by_func); */
      /*   printf("  by_Beineke  %d\n", by_Beineke); */
      /*   printf("%s", ntog6((graph*)g,m,n)); */
      /*   nautyextra_view(g,m,n); */
      /*   if (++count_bad >= max_bad) assert(0); */
      /* } */
    }
  }

  {
    DEBUG(printf("sample linegraphs\n"));
    static struct {
      const char *str;
      int want;
    } data[] = {
      { "Cn",  1 },   /* diamond, line graph of triangle plus leaf */
      { "D^{", 0 },
      { "Dr[", 0 },
    };
    int i;
    for (i = 0; i < numberof(data); i++) {
      const char *str = data[i].str;
      int want = data[i].want;
      int n = graphsize((char*)str);
      int m = SETWORDSNEEDED(n);
      assert (m==1);
      assert (n<=5);
      setword g[5];
      stringtograph((char*)str,g,m);

      setword g2[numberof(g)];
      nautyextra_ranrelabel(g,m,n,g2);

      int got = nautyextra_is_linegraph (g2,m,n);
      if (got != want) {
        printf("%s got %d want %d\n", str, got, want);
        nautyextra_view_synchronous = 1;
        nautyextra_view(g2,m,n);
        if (++count_bad >= max_bad) assert(0);
      }
    }
  }

  {
    DEBUG(printf("nautyextra_linegraph() and nautyextra_is_linegraph()\n"));
    const int n = 4;
    const int m = SETWORDSNEEDED(n);
    setword g[m];
    nautyextra_make_cycle (g,m,n, 0);
    nautyextra_view_synchronous = 1;
    /* nautyextra_view(g,m,n); */

    const int ln = nautyextra_num_edges(g,m,n);
    const int lm = SETWORDSNEEDED(ln);
    setword lg[lm];
    nautyextra_linegraph (g,m,n, lg,lm,ln);

    assert (lm==1);
    assert (nautyextra_is_linegraph(lg,lm,ln) == TRUE);
    /* nautyextra_view(lg,lm,ln); */
  }

  {
    DEBUG(printf("nautyextra_is_clique()\n"));

    int n = 6;
    int m = SETWORDSNEEDED(n);
    DYNALLOC2(graph,g, g_sz, m,n,"malloc");

    nautyextra_make_complete (g,m,n, 0);
    assert(loopcount(g,m,n) == 0);
    /* nautyextra_view(g,m,n); */

    setword ss = 0;
    assert (nautyextra_is_clique(g,m,n, &ss) == TRUE);
    ADDELEMENT(&ss,0);
    assert (nautyextra_is_clique(g,m,n, &ss) == TRUE);
    ADDELEMENT(&ss,3);
    ADDELEMENT(&ss,4);
    assert (nautyextra_is_clique(g,m,n, &ss) == TRUE);

    assert (nautyextra_clique_number(g,m,n) == n);
  }
  {
    /* nautyextra_is_clique() */
    int n = 6;
    int m = SETWORDSNEEDED(n);
    DYNALLOC2(graph,g, g_sz, m,n,"malloc");

    nautyextra_make_path (g,m,n, 0);
    assert(loopcount(g,m,n) == 0);
    /* nautyextra_view(g,m,n); */

    setword ss = 0;
    assert (nautyextra_is_clique(g,m,n, &ss) == TRUE);   /* empty */
    ADDELEMENT(&ss,0);
    ADDELEMENT(&ss,1);
    assert (nautyextra_is_clique(g,m,n, &ss) == TRUE);   /* pair */
    ADDELEMENT(&ss,2);
    assert (nautyextra_is_clique(g,m,n, &ss) == FALSE);  /* path-3 not clique */

    assert (nautyextra_clique_number(g,m,n) == 2);
  }

  {
    /* most indomsets */
    int bad = 0;

    int n;
    for (n=6; n<=20; n++) {
      int m = SETWORDSNEEDED(n);
      DYNALLOC2(graph,g, g_sz, m,n,"malloc");
      nautyextra_make_most_indomsets (g,m,n);
      nautyextra_validate(g,m,n);
      nauty_to_sg(g,&sg,m,n);

      struct nautyextra_indomnum_and_count info;
      nautyextra_indomnum_and_count_sg(&sg,&info);
      unsigned long long got = nautyextra_indomsets_count_sg(&sg);
      unsigned long long want = test_most_indomsets(n);
      /* printf("n=%d sets %Lu indomnum %d count %Lu\n", n, got, */
      /*        info.indomnum, info.count); */
      if (got != want || got != info.count) {
        bad++;
        printf("nautyextra_make_most_indomsets() n=%d sets got %Lu want %Lu info.count %Lu\n",
               n, got, want, info.count);
        nautyextra_view_synchronous = 1;
        nautyextra_view(g,m,n);
      }
    }
    if (bad) {
      if (++count_bad >= max_bad) assert(0);
    }
  }
  {
    /* Mobius ladder domsets count per A284663
     */
    int bad = 0;
    static const unsigned long long A284663[] = {
      3, 15, 51, 179, 663, 2439, 8935, 32771, 120219, 440975, 1617531,
      /* 5933267, 21763823, 79831879, 292831311, 1074134531, 3940032883 */
    };
    int i;
    for (i=0; i<numberof(A284663); i++) {
      int N = i+1;
      DYNALLSTAT(graph,g, g_sz);
      int n = N<<1;
      int m = SETWORDSNEEDED(n);
      DYNALLOC2(graph,g, g_sz, m,n,"malloc");
      nautyextra_make_Mobius_ladder (g,m,n);
      nautyextra_validate(g,m,n);
      nauty_to_sg(g,&sg,m,n);

      unsigned long long got = nautyextra_domsets_count_sg(&sg);
      unsigned long long want = A284663[i];
      if (got != want) {
        bad++;
        printf("N=%d got %Lu want %Lu\n", N, got, want);
      }
    }
    if (bad) {
      if (++count_bad >= max_bad) assert(0);
    }
  }

  {
    /* nautyextra_indomnum_vpar() test case
     */
    /*                          0, 1  2  3  4  5  6  */
    static const int vpar[] = { 0, 0, 1, 1, 3, 3, 3 };
    int n = numberof(vpar) - 1;

    struct nautyextra_indomnum_and_count  info_vpar;
    nautyextra_indomnum_and_count_vpar (vpar,n, &info_vpar);
    int indomnum = nautyextra_indomnum_vpar (vpar,n);

    if (info_vpar.indomnum != indomnum) {
      printf("sample n=%d indomnum_and_count by_vpar %d %Lu  indiv %d\n",
             n, info_vpar.indomnum, info_vpar.count,
             indomnum);
      nautyextra_print_vpar(vpar,n);
      if (++count_bad >= max_bad) assert(0);
    }
  }

  {
    /* Jou and Lin noduplicated leaf extremal indnum */
    const int max_n = 20;
    int vpar[max_n+1];
    int n;
    for (n=4; n<=20; n++) {
      make_noduplicated_leaf_extremal_indnum(vpar,n);
      /* nautyextra_view_vpar(vpar,n); */
      assert (! nautyextra_any_duplicated_leaf_vpar(vpar,n));

      int indnum_by_formula = (2*n-1)/3;
      assert(nautyextra_indnum_vpar(vpar,n) == indnum_by_formula);
    }
  }

  {
    /* nautyextra_vpar_to_g6() of empty graph */
    static const char graph6_empty_graph[] = "?\n";
    const char *g6 = nautyextra_vpar_to_g6(NULL,0);
    /* puts(g6); */
    assert (strcmp(g6, graph6_empty_graph) == 0);
  }
  {
    /* nautyextra_vpar_to_s6() of empty graph */
    static const char sparse6_empty_graph[] = ":?\n";
    const char *s6 = nautyextra_vpar_to_s6(NULL,0);
    /* puts(s6); */
    assert (strcmp(s6, sparse6_empty_graph) == 0);
  }
  {
    /* nautyextra_vpar_to_d6() of empty graph */
    static const char digraph6_empty_graph[] = "&?\n";
    int directed;
    for (directed = 0; directed <= 1; directed++) {
      const char *d6 = nautyextra_vpar_to_d6(NULL,0, directed);
      assert (strcmp(d6, digraph6_empty_graph) == 0);
    }
  }
  {
    /* nautyextra_vpar_to_d6() of 1 vertex */
    static const char digraph6_empty_graph[] = "&@?\n";
    const int n = 1;
    static const int vpar[2] = { 0, 0 };
    int directed;
    for (directed = 0; directed <= 1; directed++) {
      const char *d6 = nautyextra_vpar_to_d6(vpar,n, directed);
      assert (strcmp(d6, digraph6_empty_graph) == 0);
    }
  }
  {
    /* sgtod6() of directed sparsegraph, preserves directedness */
    {
      sg.nv = 0;
      nautyextra_add_arc_sg(&sg,0,1);
      const char *d6 = sgtod6(&sg);
      /* printf("\"%s\"\n", d6); */
      assert (strcmp(d6, "&AO\n") == 0);
    }
    {
      sg.nv = 0;
      nautyextra_add_arc_sg(&sg,1,0);
      const char *d6 = sgtod6(&sg);
      /* printf("\"%s\"\n", d6); */
      assert (strcmp(d6, "&AG\n") == 0);
    }
  }

  {
    volatile int n = 0;
    int m = SETWORDSNEEDED(n);
    if (m != 0) {
      printf ("SETWORDSNEEDED(0) = %d\n", m);
      if (++count_bad >= max_bad) assert(0);
    }

    setword ss[1];
    EMPTYSET(ss,m);
  }
  {
    /* sample 0 --> 1
       vs     0 <-- 1
    */
    int n = 2;
    int m = SETWORDSNEEDED(n);
    setword g1[m],g2[m];
    EMPTYGRAPH(g1,m,n);
    EMPTYGRAPH(g2,m,n);
    ADDONEARC(g1, 0,1, m);
    ADDONEARC(g2, 1,0, m);
    assert (nautyextra_num_arcs(g1,m,n) == 1);
    assert (nautyextra_num_arcs(g2,m,n) == 1);
    assert (nautyextra_is_isomorphic(g1,g2,m,n) == TRUE);
  }
  {
    /* sample 0 --> 1 --> 2
       vs     0 --> 1 <-- 2
    */
    int n = 3;
    int m = SETWORDSNEEDED(n);
    setword g1[m],g2[m];
    EMPTYGRAPH(g1,m,n);
    EMPTYGRAPH(g2,m,n);
    ADDONEARC(g1, 0,1, m); ADDONEARC(g1, 1,2, m);
    ADDONEARC(g2, 0,1, m); ADDONEARC(g2, 2,1, m);

    assert (nautyextra_num_arcs(g1,m,n) == 2);
    assert (nautyextra_num_arcs(g2,m,n) == 2);
    assert (nautyextra_is_isomorphic(g1,g2,m,n) == FALSE);

    nautyextra_to_undirected(g1,m,n);
    nautyextra_to_undirected(g2,m,n);
    assert (nautyextra_is_isomorphic(g1,g2,m,n) == TRUE);
    assert (nautyextra_num_arcs(g1,m,n) == 4);
    assert (nautyextra_num_arcs(g2,m,n) == 4);
  }
  {
    assert (nautyextra_binomial(0,3) == 0);
    assert (nautyextra_binomial(2,100) == 0);
    assert (nautyextra_binomial(2,1) == 2);
    assert (nautyextra_binomial(3,1) == 3);
    assert (nautyextra_binomial(3,2) == 3);
    assert (nautyextra_binomial(4,0) == 1);
    assert (nautyextra_binomial(4,1) == 4);
    assert (nautyextra_binomial(4,2) == 6);
    assert (nautyextra_binomial(4,3) == 4);
    assert (nautyextra_binomial(4,4) == 1);
  }

  {
    /* nautyextra_diameter_info_vpar() samples
       vpar_view([4, 1, 1, 0, 1])
    */
    static const struct {
      int n;
      int vpar[20];
      int want_diameter, want_endpoints, want_vertices;
    } data[] = {
      { 10, {0,  2, 7, 5, 10, 9, 7, 10, 10, 1, 0 },  7, 3, 9 },
      { 10, {0,  6,4,9,6,1,0,1,9,6,4 },              4, 6, 10 },
      {  9, {0,  7,9,5,5,6,1,0,7,1 },                5, 4, 9 },
      {  5, {0,  4, 1, 1, 0, 1 },                    2, 4, 5 },
      { 10, {0,  9, 6, 9, 7, 8, 1, 9, 0, 5, 6 },     5, 4, 9 },
      { 14, {0,  0,1,2,2,2,2,2,2,8,9,10,11,12,12 },  7, 8, 14 },
      { 10, {0,  4, 1, 7, 10, 0, 5, 6, 6, 7, 5 },    7, 3, 9 }
    };
    int i;
    for (i = 0; i < numberof(data); i++) {
      int n = data[i].n;
      const int *vpar = data[i].vpar;
      const int want_diameter  = data[i].want_diameter;
      const int want_endpoints = data[i].want_endpoints;
      const int want_vertices  = data[i].want_vertices;

      int endpoints_by_ecc = test_diameter_endpoints_by_eccentricity(vpar,n);
      assert (endpoints_by_ecc == want_endpoints);

      int vertices_by_paths = test_diameter_vertices_by_paths(vpar,n);
      assert (vertices_by_paths == data[i].want_vertices);

      struct nautyextra_diameter_info  info;
      nautyextra_diameter_info_vpar (vpar,n, &info);
      assert (info.diameter  == want_diameter);

      if (info.num_endpoints != want_endpoints) {
        printf ("nautyextra_diameter_info_vpar() endpoints\n");
        printf ("  want  %d\n", want_endpoints);
        printf ("  got   %d\n", info.num_endpoints);
        nautyextra_view_synchronous = 1;
        nautyextra_view_vpar(vpar,n);
        if (++count_bad >= max_bad) assert(0);
      }
      if (info.num_vertices != want_vertices) {
        printf ("nautyextra_diameter_info_vpar() vertices\n");
        printf ("  want  %d\n", want_vertices);
        printf ("  got   %d\n", info.num_vertices);
        nautyextra_view_synchronous = 1;
        nautyextra_view_vpar(vpar,n);
        if (++count_bad >= max_bad) assert(0);
      }
    }
  }
  {
    /* nautyextra_make_star_vpar() */
    const int n_max = 10;
    int vpar[n_max+1];
    int n;
    for (n=5; n<=n_max; n++) {
      nautyextra_make_star_vpar (vpar,n);
      nautyextra_reroot_vpar (vpar,n, 4);
      {
        /* nautyextra_diameter_info_vpar() */
        struct nautyextra_diameter_info  info;
        nautyextra_diameter_info_vpar (vpar,n, &info);
        assert (info.diameter == (n>=3 ? 2 : n-1));
        int want_endpoints = (n<=2 ? n : n-1);
        if (info.num_endpoints != want_endpoints) {
          printf ("star-%d diameter endpoints\n", n);
          printf ("  want  %d\n", want_endpoints);
          printf ("  got   %d\n", info.num_endpoints);
          if (++count_bad >= max_bad) assert(0);
        }
      }
      {
        int got  = nautyextra_claw_count_vpar(vpar,n);
        int want = nautyextra_binomial(n-1, 3);
        if (got != want) {
          printf("make_star_vpar() n=%d count claws got %d want %d\n",
                 n,got,want);
          if (++count_bad >= max_bad) assert(0);
        }
      }
      {
        int got = nautyextra_domnum_vpar(vpar,n);
        int want = (n==0 ? 0 : 1);
        if (got != want) {
          printf("make_star_vpar() n=%d domnum got %d want %d\n",
                 n,got,want);
          if (++count_bad >= max_bad) assert(0);
        }
      }
      {
        int got  = nautyextra_limb_height_vpar(vpar,n);
        int want = (n<=1 ? 0 : 1);
        if (got != want) {
          printf("star n=%d nautyextra_limb_height_vpar() got %d want %d\n",
                 n,got,want);
          if (++count_bad >= max_bad) assert(0);
        }
      }
    }
  }

  {
    /* nautyextra_claw_count() on bipartite */
    int a, b;
    for (a = 0; a <= 6; a++) {
      for (b = 0; b <= 6; b++) {
        int n = a + b;
        int m = SETWORDSNEEDED(n);
        DYNALLOC2(graph,g,g_sz,m,n,"malloc");
        nautyextra_make_complete_bipartite(g,m,n, a,b);
        assert (nautyextra_is_undirected(g,m,n) == TRUE);
        assert (loopcount(g,m,n) == 0);
        nautyextra_validate(g,m,n);

        NAUTYEXTRA_TYPE_ULL got  = nautyextra_claw_count(g,m,n);
        NAUTYEXTRA_TYPE_ULL want
          = a*nautyextra_binomial(b,3) + b*nautyextra_binomial(a,3);
        if (got != want) {
          printf("nautyextra_claw_count() on bipartite %d,%d\n", a,b);
          printf("  got  %Lu\n", (unsigned long long) got);
          printf("  want %Lu\n", (unsigned long long) want);
          nautyextra_view_synchronous = 1;
          nautyextra_view(g,m,n);
          if (++count_bad >= max_bad) assert(0);
        }
      }
    }
  }
  {
    /* Some bi-star subdivision minimal domsets counts.
     */
    static const struct {
      int a, b, length;
      unsigned long long count;
    } data[] = {
      { 1, 1, 1,     2 },
      { 4, 4, 1,     4 },
      { 2, 2, 2,     9 },   /* path 7 *-*-@-*-@-*-* */
      { 5, 5, 2,   737 },
      { 6, 4, 2,   729 },
      { 7, 7, 2, 12161 }
    };
    const int max_n = 128;
    int vpar[max_n+1];
    int i;
    for (i = 0; i < numberof(data); i++) {
      int n = nautyextra_make_bistar_n (data[i].a, data[i].b, data[i].length);
      assert (n <= max_n);
      nautyextra_make_bistar_vpar (vpar, data[i].a, data[i].b, data[i].length);
      /* nautyextra_print_vpar (vpar,n); */
      nautyextra_validate_vpar(vpar,n);
      unsigned long long  count_by_vpar
        = nautyextra_minimal_domsets_count_vpar(vpar,n);
      if (count_by_vpar != data[i].count) {
        printf("bi-star %d,%d,%d nautyextra_minimal_domsets_count_vpar()\n",
               data[i].a, data[i].b, data[i].length);
        printf("  count_by_vpar  %Lu\n", count_by_vpar);
        printf("  data           %Lu\n", data[i].count);
        if (++count_bad >= max_bad) assert(0);
      }
    }
  }

  {
    /* nautyextra_path_vertices_vpar
       1-->2-->3-->4-->5<--6
    */
    int vpar[20]; memset(vpar,'\0',sizeof(vpar));
    int n = 6;
    nautyextra_make_path_vpar (vpar,n);
    nautyextra_reroot_vpar (vpar,n, 5);
    int path[20]; memset(path,'\0',sizeof(path));
    int path_length = nautyextra_path_vertices_vpar (vpar, 1,6, path);
    if (path_length != 5)
      printf("path_length %d\n", path_length);
    assert(path_length == 5);

    static const int want[20] = { 1,2,3,4,5,6 };
    if (memcmp(path,want,sizeof(vpar))) {
      int i;
      for (i=0; i<=path_length; i++)
        printf(" %d", path[i]);
      printf("\n");
      assert(memcmp(path,want,sizeof(vpar)) == 0);
    }
    {
      int got  = nautyextra_limb_length_vpar(vpar,n);
      int want = (n<=1 ? 0 : n-1);
      if (got != want) {
        printf("path n=%d nautyextra_limb_length_vpar() got %d want %d\n",
               n,got,want);
        nautyextra_view_vpar(vpar,n);
        if (++count_bad >= max_bad) assert(0);
      }
    }
  }

  {
    /* nautyextra_matchnum_and_count_vpar() test case

       .  4---3---2\        4-3 2-1 5-6     4 matchings
       .            1       4-3 2-1   6-7
       .  7---6---5/        4-3   1-5 6-7
       .                      3-2 1-5 6-7

       at     5 with=1[1]
       parent 1 with=2[1] without=1[2]
    */
    /*                          0, 1  2  3  4  5  6  7  */
    static const int vpar[] = { 0, 0, 1, 2, 3, 1, 5, 6 };
    int n = numberof(vpar) - 1;

    struct nautyextra_matchnum_and_count  info_sets, info_vpar;
    test_matchnum_and_count_vpar_by_sets (vpar,n, &info_sets);
    nautyextra_matchnum_and_count_vpar (vpar,n, &info_vpar);

    if (info_vpar.matchnum != info_sets.matchnum
        || info_vpar.count != info_sets.count) {
      printf("samples n=%d matchnum_and_count by_vpar %d %Lu by sets %d %Lu\n",
             n, info_vpar.matchnum, info_vpar.count,
             info_sets.matchnum, info_sets.count);
      nautyextra_print_vpar(vpar,n);
      if (++count_bad >= max_bad) assert(0);
    }
  }



  {
    /* nautyextra_path_centres_vpar() */
    /*                          0, 1  2  3  4  5  6  7  */
    static const int vpar[] = { 0, 0, 1, 2, 3, 4, 5, 6 };
    /* int n = numberof(vpar) - 1; */
    {
      int c1 = -1, c2 = -1;
      nautyextra_path_centres_vpar(vpar, 6,3,3, &c1,&c2);
      assert (c1 == 5 && c2 == 4);
    }
    {
      int c1 = -1, c2 = -1;
      nautyextra_path_centres_vpar(vpar, 3,6,3, &c1,&c2);
      assert (c1 == 4 && c2 == 5);
    }
    {
      int c1 = -1, c2 = -1;
      nautyextra_path_centres_vpar(vpar, 7,3,3, &c1,&c2);
      assert (c1 == 5 && c2 == 5);
    }
    {
      int c1 = -1, c2 = -1;
      nautyextra_path_centres_vpar(vpar, 1,5,1, &c1,&c2);
      assert (c1 == 3 && c2 == 3);
    }
  }
  {
    /* nautyextra_path_centres_vpar() */
    /*                          0, 1  2  3  4  5  6  7  8 */
    static const int vpar[] = { 0, 2, 3, 4, 5, 0, 5, 6, 7 };
    /* int n = numberof(vpar); */
    {
      int c1 = -1, c2 = -1;
      nautyextra_path_centres_vpar(vpar, 1,8,5, &c1,&c2);
      assert (c1 == 4 && c2 == 5);
    }
    {
      int c1 = -1, c2 = -1;
      nautyextra_path_centres_vpar(vpar, 6,1,5, &c1,&c2);
      assert (c1 == 4 && c2 == 3);
    }
  }

  {
    /* nautyextra_make_hypercube() domsets count */
    static const unsigned long long want_domsets[] = {
      1,3,11,183,45707,  /* A284704 */
    };
    int d;
    for (d = 0; d < numberof(want_domsets); d++) {
      int n = 1 << d;
      int m = SETWORDSNEEDED(n);
      DYNALLOC2(graph,g,g_sz,m,n,"malloc");
      nautyextra_make_hypercube(g,m,n, d);
      nauty_to_sg (g,&sg,m,n);
      nautyextra_validate_sg(&sg);

      unsigned long long got = nautyextra_domsets_count_sg(&sg);
      if (got != want_domsets[d]) {
        printf("hypercube %d domnum got %Lu want %Lu\n",
               d, got, want_domsets[d]);
        if (++count_bad >= max_bad) assert(0);
      }
    }
  }
  {
    /* nautyextra_make_hypercube() domnum */
    static const int want_domnum[] = {
      1,
      1,
      2, 2, 4  /* , 4, 8, 14, 24 */      /* A247181 */
    };
    int d;
    for (d = 0; d < numberof(want_domnum); d++) {
      int n = 1 << d;
      int m = SETWORDSNEEDED(n);
      DYNALLOC2(graph,g,g_sz,m,n,"malloc");
      nautyextra_make_hypercube(g,m,n, d);
      nautyextra_validate(g,m,n);
      assert (nautyextra_is_undirected(g,m,n) == TRUE);
      nauty_to_sg (g,&sg,m,n);
      nautyextra_validate_sg(&sg);

      struct nautyextra_domnum_and_count info;
      nautyextra_domnum_and_count_sg(&sg, &info);
      if (info.domnum != want_domnum[d]) {
        printf("hypercube %d domnum got %d want %d\n",
               d, info.domnum, want_domnum[d]);
        if (++count_bad >= max_bad) assert(0);
      }
    }
  }

  {
    /* nautyextra_claw_count() */
    static const char *data[] = {   /* these containing claws */
      ":Do@G~",
      ":DkGCb",
      ":DkGCn",
      ":Dk@_QN",
      ":Dk@_Q",
      ":Dk@I@G~",
      ":Dk@I@J",
      ":CCKS^",  /* claw with self-loops everywhere */
    };
    int i;
    for (i = 0; i < numberof(data); i++) {
      const char *str = data[i];
      graph g[8];
      int n = graphsize((char*)str);
      int m = SETWORDSNEEDED(n);
      assert (m == 1);
      assert (n <= numberof(g));
      stringtograph((char*)str, g,m);

      graph g2[numberof(g)];
      nautyextra_ranrelabel(g,m,n,g2);
      int claws = nautyextra_claw_count(g2,m,n);

      if (! (claws >= 1)) {
        nautyextra_view_synchronous = 1;
        nautyextra_view(g2,m,n);
      }
      assert (claws >= 1);
    }
  }
  {
    /* claw */
    static const int vpar[] = { 0, 0,1,1,1 };
    const int n = 4;
    int claws = nautyextra_claw_count_vpar(vpar,n);
    assert (claws == 1);
  }


  {
    /* complete binary tree by vpar */

    const int max_h = 6;
    const int vpar_size = (1 << max_h);
    int vpar[vpar_size];
    int h;
    for (h = 0; h <= max_h; h++) {
      int n = nautyextra_make_complete_tree_size (2,h);
      assert (n < vpar_size);
      nautyextra_make_complete_tree_vpar (vpar,2,h);

#if 0
      {
        /* nautyextra_domnum_vpar() */
        int by_vpar    = nautyextra_domnum_vpar(vpar,n);
        int by_formula = 0;
        int i;
        for (i = 0; i < h; i+=2) by_formula += 1<<i;
        if (by_vpar != by_formula) {
          nautyextra_print_vpar(vpar,n);
          if (++count_bad >= max_bad) assert(0);
        }
      }
#endif
      {
        /* nautyextra_matchnum_vpar() */
        int by_vpar  = nautyextra_matchnum_vpar (vpar, n);
        int by_formula = 0;
        int i;
        for (i = h-2; i >= 0; i-=2) by_formula += 1<<i;
        if (by_formula != by_vpar) {
          printf("nautyextra_matchnum_vpar() h=%d  by_vpar %d by_formula %d\n",
                 h, by_vpar, by_formula);
          if (++count_bad >= max_bad) assert(0);
        }
      }
    }
  }

  {
    /* nautyextra_make_cycle_sg() edges */
    /*                              0  1  2  3  4 */
    static const int want_nde[] = { 0, 0, 2, 6, 8 };
    static const int max_n = numberof(want_nde) - 1;
    int n;
    for (n = 0; n <= max_n; n++) {
      nautyextra_make_cycle_sg (&sg,n);
      /* nautyextra_print_sg(&sg); */
      assert (nautyextra_is_undirected_sg(&sg) == TRUE);
      nautyextra_validate_sg(&sg);
      if (sg.nde != want_nde[n]) {
        printf("nautyextra_make_cycle_sg() n=%d  nde=%d want %d\n",
               n, sg.nde, want_nde[n]);
        if (++count_bad >= max_bad) assert(0);
      }
    }
  }

  /* { */
  /*   static const int vpar[] = { 0,0,1,2,3,4,5 }; */
  /*   nautyextra_indomsets_count_vpar(vpar,2); */
  /*   /\* return 0; *\/ */
  /* } */
  /* { */
  /*   static const int vpar[] = { 0,0,1,2,3,4,5 }; */
  /*   nautyextra_indomsets_count_vpar(vpar,3); */
  /*   /\* return 0; *\/ */
  /* } */
  /* { */
  /*   static const int vpar[] = { 0,0,1,2,3,4,5 }; */
  /*   nautyextra_indomsets_count_vpar(vpar,4); */
  /*   /\* return 0; *\/ */
  /* } */

  {
    /* nautyextra_vpar_to_sg() */
    static const int vpar[] = {0, 0,1 };
    int directed;
    for (directed = 0; directed <= 2; directed++) {
      nautyextra_vpar_to_sg(vpar,0, &sg, directed);
      nautyextra_validate_sg(&sg);

      nautyextra_vpar_to_sg(vpar,1, &sg, directed);
      nautyextra_validate_sg(&sg);

      nautyextra_vpar_to_sg(vpar,2, &sg, directed);
      nautyextra_validate_sg(&sg);
    }
  }

  if (0) {
    /* Tree T1 from
       Marcin Krzywkowski, "Trees Having Many Minimal Dominating Sets"

       sqrt(2)
       41
       GP-Test  41^2 == 1681
       41^3
       GP-Test  9^2 == 81
       GP-Test  9^3 == 729
    */
    int w;
    for (w = 6; w <= 6; w++) {
      int k;
      for (k = 1; k <= 1; k++) {
        test_make_T1(&sg,w,k);
        assert (nautyextra_is_undirected_sg(&sg) == TRUE);
        /* nautyextra_view_sg(&sg); */

        nautyextra_want_progress = 1;
        int count_dom = 0;
        int count_minimal_dom = 0;
        int n = sg.nv;
        setword ss;
        NAUTYEXTRA_SET1_LOOP
          (ss,n,
           int is_dom         = nautyextra_is_domset_sg (&sg,&ss);
           int is_minimal_dom = nautyextra_is_minimal_domset_sg (&sg,&ss);
           count_dom         += is_dom;
           count_minimal_dom += is_minimal_dom;

           /* if (is_minimal_dom) { */
           /*   printf("%s\n", nautyextra_set_str(&ss,n)); */
           /* } */
           /* if (++x > 2) exit (0); */

           if ((ss & 0xFF) == 0 && nautyextra_progress())
             printf("%d %d 0x%0*X\r", count_dom, count_minimal_dom, (27+3)/4, ss);
           );
        printf("end                 \n");

        printf("T1 w=%d k=%d n=%d  domsets %d minimals %d\n",
               w, k, n, count_dom, count_minimal_dom);

        if (k==1 && w==6) {
          assert (count_minimal_dom == 12161);
          assert (sg.nv == 27);
        }
      }
    }
    nautyextra_want_progress = 0;
  }
  {
    /* nautyextra_make_star() */
    const int n_max = 10;
    DYNALLSTAT(graph,g,g_sz);
    int n;
    for (n=0; n<=n_max; n++) {
      int m = SETWORDSNEEDED(n);
      DYNALLOC2(graph,g,g_sz,m,n,"malloc");
      nautyextra_make_star(g,m,n, 0);
      assert (nautyextra_is_undirected(g,m,n) == TRUE);
      nautyextra_validate(g,m,n);
      /* if (n==5) nautyextra_view(g,m,n); */
      {
        int got  = loopcount(g,m,n);
        int want = 0;
        if (got != want) {
          printf("make_star() n=%d loopcount got %d want %d\n", n,got,want);
          if (++count_bad >= max_bad) assert(0);
        }
      }
      {
        int got  = nautyextra_claw_count(g,m,n);
        int want = nautyextra_binomial(n-1,3);
        if (got != want) {
          printf("make_star() n=%d count claws got %d want %d\n",
                 n,got,want);
          if (++count_bad >= max_bad) assert(0);
        }
      }
    }
  }

  {
    sg.nv = sg.nde = 0;
    nautyextra_ensure_vertex_sg(&sg,1);
    assert (strcmp(sgtos6(&sg),":A\n")==0);

    nautyextra_add_edge_sg(&sg,0,1);
    assert (sg.d[0] == 1);
    assert (sg.d[1] == 1);
    printf("%s",sgtos6(&sg));
    assert (strcmp(sgtos6(&sg),":An\n")==0);

    nautyextra_add_edge_sg(&sg,0,1);
    assert (sg.d[0] == 2);
    assert (sg.d[1] == 2);
    assert (strcmp(sgtos6(&sg),":Ab\n")==0);

    nautyextra_add_edge_sg(&sg,0,1);
    assert (sg.d[0] == 3);
    assert (sg.d[1] == 3);
    /* printf("%s", sgtos6(&sg)); */
    assert (strcmp(sgtos6(&sg),":A_\n")==0);
  }

  {
    /* path graphs of various lengths n */
    const int n_max = 70;
    static const int n_list[] = { 0,1,2,3,4,5,6,7,8,
                                  29,30,31,32,33,34,40 };
    int m_max = SETWORDSNEEDED(n_max);
    set gspace[n_max][m_max];
    graph *g = (graph*)gspace;
    int i;
    for (i = 0; i < numberof(n_list); i++) {
      int n = n_list[i];
      assert(n<=n_max);
      int m = SETWORDSNEEDED(n);
      nautyextra_make_path (g,m,n, 0);
      assert (nautyextra_is_undirected(g,m,n) == TRUE);
      nautyextra_validate(g,m,n);

      {
        int got  = nautyextra_claw_count(g,m,n);
        int want = 0;
        if (got != want) {
          printf("nautyextra_make_path() n=%d count claws got %d want %d\n",
                 n,got,want);
          if (++count_bad >= max_bad) assert(0);
        }
      }
      {
        /* nautyextra_indnum_and_count() */
        struct nautyextra_indnum_and_count  info;
        nautyextra_indnum_and_count(g,m,n,&info);

        assert (info.indnum >= 0);
        int want_indnum = test_path_indnum(n);
        unsigned long long want_count = test_path_indnum_count(n);
        if (info.indnum != want_indnum
            || info.count != want_count) {
          printf("nautyextra_make_path() n=%d nautyextra_indnum_and_count() indnum=%d count=%Lu want %d %Lu\n",
                 n, info.indnum, info.count, want_indnum, want_count);
          nautyextra_print_matrix(g,m,n);
          if (++count_bad >= max_bad) assert(0);
        }
      }
    }
  }
}

static void
try_nautyextra_indomsets_count_tree_max (void)
{
  {
    /* nautyextra_make_star_arms2_vpar()
       attains nautyextra_indomsets_count_tree_max() */
    const int n_max = 100;
    int vpar[n_max+1];
    int n;
    for (n=0; n<=n_max; n++) {
      unsigned long long want = nautyextra_indomsets_count_tree_max(n);

      nautyextra_make_star_arms2_vpar (vpar, n);
      /* nautyextra_print_vpar (vpar,n); */
      /* nautyextra_view_vpar(vpar,n); */

      unsigned long long got = nautyextra_indomsets_count_vpar(vpar,n);
      if (got != want) {
        printf("make_star_arms2_vpar() not nautyextra_indomsets_count_tree_max() n=%d got %Lu want %Lu\n",
               n,got,want);
        if (++count_bad >= max_bad) assert(0);
      }
    }
  }
  {
    /* Wilf, A228693
       f(n) = if(n==0,1, n%2==0,2^(n/2-1)+1, 2^((n-1)/2));
       vector(10,n,n--; f(n))
    */
    static const unsigned long long samples[] = {
      1, 1, 2, 2, 3, 4, 5, 8, 9, 16, 17, 32, 33,
      64, 65, 128, 129, 256, 257
    };
    int n;
    for (n = 0; n < numberof(samples); n++) {
      unsigned long long got = nautyextra_indomsets_count_tree_max(n);
      unsigned long long want = samples[n];
      if (got != want) {
        printf("nautyextra_indomsets_count_tree_max() n=%d got %Lu want %Lu\n",
               n,got,want);
        if (++count_bad >= max_bad) assert(0);
      }
    }
  }
}

int
main (int argc, char **argv)
{
  setbuf(stdout,NULL);
  SG_INIT(sg);   /* globals */
  SG_INIT(sg2);
  SG_INIT(canon_sg);
  SG_INIT(canon_sg2);

  {
    int i;
    for (i = 1; i < argc; i++) {
      if (strcmp(argv[i], "-v") == 0) {
        option_verbose++;
      } else {
        printf ("unrecognised option: %s\n", argv[1]);
      }
    }
  }

  if (1) {
    printf ("random by time\n");
    INITRANBYTIME;
  } else {
    printf ("fixed random\n");
  }
  printf ("KRAN %ld", KRAN(1000));
  printf (" %ld", KRAN(1000));
  printf (" %ld", KRAN(1000));
  printf ("\n");

  {
    int n = tree_max_vertices;
    if (n < graph_max_vertices) n = graph_max_vertices;
    int m = SETWORDSNEEDED(n);
    nauty_check(WORDSIZE,m,n,NAUTYVERSIONID);
  }

  try_nautyextra_indomsets_count_tree_max();
  try_various();
  try_gentreeg();
  try_geng();
  try_geng_unconnected();


#if 0
  struct mallinfo m = mallinfo();
  printf ("mallinfo arena %dk mmap %dk\n", m.arena/1000, m.hblkhd/1000);
#endif

  printf("tests total %d\n", count_total);
  return (count_bad ? 1 : 0);
}
